#!/usr/bin/env python

#******************************************************************************
#
#  rpnDice.py
#
#  rpnChilada dice simulation operators
#  copyright (c) 2021, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import itertools
import re

from collections import Counter
from random import randrange

from mpmath import arange

from rpn.rpnDebug import debugPrint
from rpn.rpnGenerator import RPNGenerator
from rpn.rpnUtils import oneArgFunctionEvaluator, twoArgFunctionEvaluator
from rpn.rpnValidator import argValidator, IntValidator, StringValidator


# write a function to enumerate the number of outcomes for a dice expression
# without actually permuting them.  i.e., rpn <dice_expr> permute_dice occurrences
#
# Is that possible?  It's definitely possible without including x and h.  With them
# it becomes a problem that's harder than I can solve, probably.
#
# c:\>rpn [ 1 6 dup ] 3 poly_power 3 enumerate -s1
# [
#     [ 3, 1 ],
#     [ 4, 3 ],
#     [ 5, 6 ],
#     [ 6, 10 ],
#     [ 7, 15 ],
#     [ 8, 21 ],
#     [ 9, 25 ],
#     [ 10, 27 ],
#     [ 11, 27 ],
#     [ 12, 25 ],
#     [ 13, 21 ],
#     [ 14, 15 ],
#     [ 15, 10 ],
#     [ 16, 6 ],
#     [ 17, 3 ],
#     [ 18, 1 ]
# ]
#
# c:\>rpn 3d6 permute_dice occurrences -s1
# [
#     [ 3, 1 ],
#     [ 4, 3 ],
#     [ 5, 6 ],
#     [ 6, 10 ],
#     [ 7, 15 ],
#     [ 8, 21 ],
#     [ 9, 25 ],
#     [ 10, 27 ],
#     [ 11, 27 ],
#     [ 12, 25 ],
#     [ 13, 21 ],
#     [ 14, 15 ],
#     [ 15, 10 ],
#     [ 16, 6 ],
#     [ 17, 3 ],
#     [ 18, 1 ]
# ]
#
# Make a new data type (dict) and then write an operator for summing multiple dicts.  Piece o' cake.
# Or use Counter...
#


#******************************************************************************
#
#  rollDice
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ StringValidator( ) ] )
def rollDiceOperator( expression ):
    values, modifier = evaluateDiceExpression( parseDiceExpression( expression ) )
    return sum( values ) + modifier


#******************************************************************************
#
#  rollMultipleDice
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ StringValidator( ), IntValidator( 0 ) ] )
def rollMultipleDice( expression, times ):
    dice = parseDiceExpression( expression )

    for _ in arange( 0, times ):
        values, modifier = evaluateDiceExpression( dice )
        yield sum( values ) + modifier


def rollMultipleDiceOperator( n, k ):
    return RPNGenerator( rollMultipleDice( n, k ) )


#******************************************************************************
#
#  enumerateDice
#
#******************************************************************************

def enumerateDice( expression ):
    return evaluateDiceExpression( parseDiceExpression( expression ), False )[ 0 ]


@oneArgFunctionEvaluator( )
@argValidator( [ StringValidator( ) ] )
def enumerateDiceOperator( n ):
    return RPNGenerator( enumerateDice( n ) )


#******************************************************************************
#
#  enumerateMultipleDice
#
#******************************************************************************

def enumerateMultipleDice( expression, count ):
    dice = parseDiceExpression( expression )

    for _ in arange( 0, count ):
        yield evaluateDiceExpression( dice )[ 0 ]


@twoArgFunctionEvaluator( )
@argValidator( [ StringValidator( ), IntValidator( 0 ) ] )
def enumerateMultipleDiceOperator( n, k ):
    return RPNGenerator( enumerateMultipleDice( n, k ) )


#******************************************************************************
#
#  permuteDice
#
#  format: [c]dv[x[p]][h[q]][-+]y
#
#  c - dice count, defaults to 1
#  v - dice value, i.e., number of sides, minumum 2
#  p - drop lowest die value(s), defaults to 1
#  q - drop highest value(s), defaults to 1
#  y - add or subtract y from the total (modifier)
#
#******************************************************************************

def permuteDice( expression ):
    dice = parseDiceExpression( expression )

    diceList = [ ]

    dropAny = False
    ranges = [ ]
    dropRanges = [ ]
    drops = [ ]
    start = 0
    modifierTotal = 0
    groupCount = len( dice )

    for diceCount, diceValue, dropLowest, dropHighest, modifier in dice:
        ranges.append( ( start, start + diceCount ) )
        start += diceCount

        if dropLowest > 0 or dropHighest > 0:
            drops.append( True )
            dropRanges.append( ( dropLowest, diceCount - dropHighest ) )
            dropAny = True
        else:
            drops.append( False )
            dropRanges.append( ( 0, 0 ) )

        modifierTotal += modifier

        for i in range( diceCount ):
            diceList.append( range( 1, diceValue + 1 ) )

    permutations = itertools.product( *diceList )

    for values in permutations:
        if dropAny:
            total = 0

            for i in range( 0, groupCount ):
                group = values[ ranges[ i ][ 0 ] : ranges[ i ][ 1 ] ]

                if drops[ i ]:
                    group = list( group )
                    group.sort( )
                    total += sum( group[ dropRanges[ i ][ 0 ] : dropRanges[ i ][ 1 ] ] )
                else:
                    total += sum( group )

                total += modifierTotal
                yield total
        else:
            yield sum( values ) + modifierTotal


@oneArgFunctionEvaluator( )
def permuteDiceOperator( n ):
    return RPNGenerator( permuteDice( n ) )


#******************************************************************************
#
#  parseDiceExpression
#
#  format: [c]dv[,[c]dv...][x[p]][h[q]][-+]y
#
#  c - dice count, defaults to 1
#  v - dice value, i.e., number of sides, minumum 2
#  p - drop lowest die value(s), defaults to 1
#  q - drop highest value(s), defaults to 1
#  y - add or subtract y from the total (modifier)
#
#  return values:
#
#  a list of 5-tuples:  dice count, dice value, drop lowest count,
#                       drop highest count, modifier
#
#******************************************************************************

def parseDiceExpression( arg ):
    counter = Counter( arg )

    if counter[ 'a' ] > 1:
        raise ValueError( 'dice expressions can only contain a single \'x\'' )

    if counter[ 'h' ] > 1:
        raise ValueError( 'dice expressions can only contain a single \'h\'' )

    if counter[ '+' ] > 1:
        raise ValueError( 'dice expressions can only contain a single \'+\'' )

    if counter[ '-' ] > 1:
        raise ValueError( 'dice expressions can only contain a single \'-\'' )

    if counter[ '+' ] + counter[ '-' ] > 1:
        raise ValueError( 'dice expressions can only have a single modifier (\'+\' or \'-\')' )

    expressions = re.split( ',', arg )

    result = [ ]

    for expr in expressions:
        pieces = re.split( '([dhx+-])', expr )

        debugPrint( 'pieces', pieces )

        diceValue = 0
        diceCount = 0
        dropLowestCount = 0
        dropHighestCount = 0
        modifier = 0

        defaultState = 0
        diceValueState = 1
        dropLowestCountState = 2
        dropHighestCountState = 3
        plusModifierState = 4
        minusModifierState = 5

        state = defaultState

        for piece in pieces:
            if state == defaultState:
                if piece == 'd':
                    state = diceValueState
                elif piece == 'x':
                    state = dropLowestCountState
                elif piece == 'h':
                    state = dropHighestCountState
                elif piece == '-':
                    state = minusModifierState
                elif piece == '+':
                    state = plusModifierState
                else:
                    if piece == '':
                        diceCount = 1
                    else:
                        diceCount = int( piece )
            elif state == diceValueState:
                diceValue = int( piece )

                if diceValue < 2:
                    raise ValueError( 'dice value must be greater than 1' )

                state = defaultState
            elif state == dropLowestCountState:
                if piece == '':
                    dropLowestCount = 1
                else:
                    dropLowestCount = int( piece )

                if dropLowestCount < 1:
                    raise ValueError( 'drop lowest count must be positive' )

                state = defaultState
            elif state == dropHighestCountState:
                if piece == '':
                    dropHighestCount = 1
                else:
                    dropHighestCount = int( piece )

                if dropHighestCount < 1:
                    raise ValueError( 'drop highest count must be positive' )

            elif state == plusModifierState:
                if piece == '':
                    modifier = 1
                else:
                    modifier = int( piece )

                state = defaultState
            elif state == minusModifierState:
                if piece == '':
                    modifier = -1
                else:
                    modifier = -int( piece )

                state = defaultState

        if diceCount == 0:
            diceCount = 1

        if dropLowestCount != 0 and diceValue == 0:
            raise ValueError( 'dice expression requires \'d\' if \'x\' is used' )

        if dropHighestCount != 0 and diceValue == 0:
            raise ValueError( 'dice expression requires \'d\' if \'h\' is used' )

        if ( dropLowestCount + dropHighestCount ) >= diceCount:
            raise ValueError( 'this dice expression is dropping as many or more dice than are being rolled' )

        debugPrint( 'expression', expr )
        debugPrint( 'diceCount', diceCount )
        debugPrint( 'diceValue', diceValue )
        debugPrint( 'dropLowestCount', dropLowestCount )
        debugPrint( 'dropHighestCount', dropHighestCount )
        debugPrint( 'modifier', modifier )

        result.append( ( diceCount, diceValue, dropLowestCount, dropHighestCount, modifier ) )

    return result


#******************************************************************************
#
#  evaluateDiceExpression
#
#  a list of 5-tuples:  dice count, dice value, drop lowest count,
#                       drop highest count, modifier
#
#******************************************************************************

def evaluateDiceExpression( args, sumIfPossible=True ):
    result = [ ]
    modifierSum = 0    # currently only a single modifier is allowed

    if sumIfPossible:
        result = [ 0 ]

    for diceCount, diceValue, dropLowestCount, dropHighestCount, modifier in args:
        modifierSum += modifier

        if dropLowestCount == 0 and dropHighestCount == 0:
            if sumIfPossible:
                for _ in range( 0, diceCount ):
                    result[ 0 ] += ( randrange( diceValue ) + 1 )
            else:
                for _ in range( 0, diceCount ):
                    result.append( randrange( diceValue ) + 1 )
        else:
            dice = [ ]

            for _ in range( 0, diceCount ):
                dice.append( randrange( diceValue ) + 1 )

            dice.sort( )
            debugPrint( 'dice', dice )

            if dropHighestCount > 0:
                debugPrint( 'drop', dice[ dropLowestCount : -dropHighestCount ] )
                result.extend( dice[ dropLowestCount : -dropHighestCount ] )
            else:
                debugPrint( 'drop', dice[ dropLowestCount : ] )
                result.extend( dice[ dropLowestCount : ] )

    return result, modifierSum


#******************************************************************************
#
#  rollSimpleDice
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ), IntValidator( 1 ) ] )
def rollSimpleDiceOperator( n, k ):
    values, modifier = evaluateDiceExpression( [ ( int( n ), int( k ), 0, 0, 0 ) ] )
    return sum( values ) + modifier



#!/usr/bin/env python

#******************************************************************************
#
#  rpnPrimeUtils.py
#
#  rpnChilada prime number utilies
#  copyright (c) 2021, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import os
import sys

from bisect import bisect_left
from pathlib import Path

from mpmath import arange, fadd, fmod, fmul, fprod, fsub

import gmpy2

from rpn.rpnDebug import debugPrint
from rpn.rpnGenerator import RPNGenerator
from rpn.rpnPrimes import primes
from rpn.rpnPersistence import cachedFunction, openPrimeCache
from rpn.rpnUtils import getUserDataPath, oneArgFunctionEvaluator, twoArgFunctionEvaluator
from rpn.rpnValidator import argValidator, IntValidator

import rpn.rpnGlobals as g


#******************************************************************************
#
#  isPrimeOperator
#
#******************************************************************************

def isPrime( n ):
    if g.zhangConjecturesAllowed and n < 1543267864443420616877677640751301:
        return isPrimeMillerRabin( int( n ) )

    if n < 3317044064679887385961981:
        return isPrimeMillerRabin( int( n ) )

    debugPrint( 'primality testing of ' + str( int( n ) ) )
    return 1 if gmpy2.is_bpsw_prp( int( n ) ) else 0


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def isCompositeOperator( n ):
    return 0 if n == 1 or isPrime( n ) else 1


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def isPrimeOperator( n ):
    return 1 if isPrime( n ) else 0


#******************************************************************************
#
#  isPrimeNumberSimple
#
#  Use this for numbers smaller than 10,000,000,000.
#
#******************************************************************************

def isPrimeNumberSimple( n ):
    # if prime is already in the list, just pick it
    if n <= max( primes ):
        i = bisect_left( primes, n )
        return i != len( primes ) and primes[ i ] == n

    # Divide by each known prime
    limit = int( n ** .5 )

    for p in primes:
        if p > limit:
            return True

        if n % p == 0:
            return False

    return True


#******************************************************************************
#
#  getNextPrimeCandidateForAny
#
#  p is the next prime candidate.  Starting with ( p - 10 ) mod 210, we only
#  need to check 43 values out of the next 210, which eliminates all multiples
#  of 2, 3, 5, and 7 and saves us a lot of unnecessary checking.
#
#  We only need to check for a prime if ( p - 10 ) mod 210 is one of the
#  following:
#
#  1, 3, 7, 9, 13, 19, 21, 27, 31, 33, 37, 43, 49, 51, 57, 61, 63, 69, 73,
#  79, 87, 91, 93, 97, 99, 103, 111, 117, 121, 127, 129, 133, 139, 141, 147,
#  153, 157, 159, 163, 169, 171, 177, 181, 183, 187, 189, 199, 201
#
#  This list can be generated by:
#
#  rpn 10 4 primorial 10 + range lambda x 5 prime is_rough filter 10 -
#
#******************************************************************************

nextPrimeModuloTable = [
    # 0
    1,
    # 1
    2, 1,
    # 3
    4, 3, 2, 1,
    # 7
    2, 1,
    # 9
    4, 3, 2, 1,
    # 13
    6, 5, 4, 3, 2, 1,
    # 19
    2, 1,
    # 21
    6, 5, 4, 3, 2, 1,
    # 27
    4, 3, 2, 1,
    # 31
    2, 1,
    # 33
    4, 3, 2, 1,
    # 37
    6, 5, 4, 3, 2, 1,
    # 43
    6, 5, 4, 3, 2, 1,
    # 49
    2, 1,
    # 51
    6, 5, 4, 3, 2, 1,
    # 57
    4, 3, 2, 1,
    # 61
    2, 1,
    # 63
    6, 5, 4, 3, 2, 1,
    # 69
    4, 3, 2, 1,
    # 73
    6, 5, 4, 3, 2, 1,
    # 79
    8, 7, 6, 5, 4, 3, 2, 1,
    # 87
    4, 3, 2, 1,
    # 91
    2, 1,
    # 93
    4, 3, 2, 1,
    # 97
    2, 1,
    # 99
    4, 3, 2, 1,
    # 103
    8, 7, 6, 5, 4, 3, 2, 1,
    # 111
    6, 5, 4, 3, 2, 1,
    # 117
    4, 3, 2, 1,
    # 121
    6, 5, 4, 3, 2, 1,
    # 127
    2, 1,
    # 129
    4, 3, 2, 1,
    # 133
    6, 5, 4, 3, 2, 1,
    # 139
    2, 1,
    # 141
    6, 5, 4, 3, 2, 1,
    # 147
    6, 5, 4, 3, 2, 1,
    # 153
    4, 3, 2, 1,
    # 157
    2, 1,
    # 159
    4, 3, 2, 1,
    # 163
    6, 5, 4, 3, 2, 1,
    # 169
    2, 1,
    # 171
    6, 5, 4, 3, 2, 1,
    # 177
    4, 3, 2, 1,
    # 181
    2, 1,
    # 183
    4, 3, 2, 1,
    # 187
    2, 1,
    # 189
    10, 9, 8, 7, 6, 5, 4, 3, 2, 1,
    # 199
    2, 1,
    # 201
    10, 9, 8, 7, 6, 5, 4, 3, 2
]


def getNextPrimeCandidateForAny( p ):
    f = int( ( p - 10 ) % 210 )
    return p + nextPrimeModuloTable[ f ]


#******************************************************************************
#
#  getPreviousPrimeCandidateForAny
#
#  p is the next prime candidate.  Starting with ( p - 10 ) mod 210, we only
#  need to check 43 values out of the next 210, which eliminates all multiples
#  of 2, 3, 5, and 7 and saves us a lot of unnecessary checking.
#
#  We only need to check for a prime if ( p - 10 ) mod 210 is one of the
#  following:
#
#  0, 2, 4, 8, 10, 14, 20, 22, 28, 32, 34, 38, 44, 50, 52, 58, 62, 64, 70, 74
#  80, 88, 92, 94, 98, 100, 104, 112, 118, 122, 128, 130, 134, 140, 142, 148
#  154, 158, 160, 164, 170, 172, 178, 182, 184, 188, 190, 200, 202,
#
#  This list can be generated by:
#
#  rpn [ 0 ] 10 4 primorial 10 + range lambda x 5 prime is_rough filter 10 - 1 + append
#
#******************************************************************************

previousPrimeModuloTable = [
    # 0
    9, 10,
    # 2
    1, 2,
    # 4
    1, 2, 3, 4,
    # 8
    1, 2,
    # 10
    1, 2, 3, 4,
    # 14
    1, 2, 3, 4, 5, 6,
    # 20
    1, 2,
    # 22
    1, 2, 3, 4, 5, 6,
    # 28
    1, 2, 3, 4,
    # 32
    1, 2,
    # 34
    1, 2, 3, 4,
    # 38
    1, 2, 3, 4, 5, 6,
    # 44
    1, 2, 3, 4, 5, 6,
    # 50
    1, 2,
    # 52
    1, 2, 3, 4, 5, 6,
    # 58
    1, 2, 3, 4,
    # 62
    1, 2,
    # 64
    1, 2, 3, 4, 5, 6,
    # 70
    1, 2, 3, 4,
    # 74
    1, 2, 3, 4, 5, 6,
    # 80
    1, 2, 3, 4, 5, 6, 7, 8,
    # 88
    1, 2, 3, 4,
    # 92
    1, 2,
    # 94
    1, 2, 3, 4,
    # 98
    1, 2,
    # 100
    1, 2, 3, 4,
    # 104
    1, 2, 3, 4, 5, 6, 7, 8,
    # 112
    1, 2, 3, 4, 5, 6,
    # 118
    1, 2, 3, 4,
    # 122
    1, 2, 3, 4, 5, 6,
    # 128
    1, 2,
    # 130
    1, 2, 3, 4,
    # 134
    1, 2, 3, 4, 5, 6,
    # 140
    1, 2,
    # 142
    1, 2, 3, 4, 5, 6,
    # 148
    1, 2, 3, 4, 5, 6,
    # 154
    1, 2, 3, 4,
    # 158
    1, 2,
    # 160
    1, 2, 3, 4,
    # 164
    1, 2, 3, 4, 5, 6,
    # 170
    1, 2,
    # 172
    1, 2, 3, 4, 5, 6,
    # 178
    1, 2, 3, 4,
    # 182
    1, 2,
    # 184
    1, 2, 3, 4,
    # 188
    1, 2,
    # 190
    1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
    # 200
    1, 2,
    # 202
    1, 2, 3, 4, 5, 6, 7, 8,
]


#******************************************************************************
#
#  getPreviousPrimeCandidateForAny
#
#******************************************************************************

def getPreviousPrimeCandidateForAny( p ):
    f = int( ( p - 10 ) % 210 )
    return p - previousPrimeModuloTable[ f ]


#******************************************************************************
#
#  getNextPrimeOperator
#
#******************************************************************************

@cachedFunction( 'next_prime' )
def getNextPrime( p, func=getNextPrimeCandidateForAny ):
    if p < 2:
        return 2

    if p == 2:
        return 3

    if p < 5:
        return 5

    if p < 7:
        return 7

    p = func( p )

    debugPrint( 'testing', p )

    while not isPrime( p ):
        p = func( p )

    return p


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def getNextPrimeOperator( n ):
    return getNextPrime( n, func=getNextPrimeCandidateForAny )


#******************************************************************************
#
#  getPreviousPrimeOperator
#
#******************************************************************************

@cachedFunction( 'previous_prime' )
def getPreviousPrime( p, func=getPreviousPrimeCandidateForAny ):
    if p < 12:
        if p < 3:
            raise ValueError( 'There is no previous prime to 2.' )

        if p == 3:
            return 2

        if p <= 5:
            return 3

        if p <= 7:
            return 5

        if p <= 11:
            return 7

    p = func( p )

    while not isPrime( p ):
        p = func( p )

    return p


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def getPreviousPrimeOperator( n ):
    return getPreviousPrime( n, func=getPreviousPrimeCandidateForAny )


#******************************************************************************
#
#  getNextPrimesOperator
#
#******************************************************************************

def getNextPrimes( p, k, func=getNextPrimeCandidateForAny ):
    result = [ ]

    for _ in arange( 0, k ):
        p = getNextPrime( p, func )
        result.append( p )

    return result


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( 0 ) ] )
def getNextPrimesOperator( n, k ):
    return getNextPrimes( n, k, func=getNextPrimeCandidateForAny )


#******************************************************************************
#
#  getPreviousPrimes
#
#******************************************************************************

def getPreviousPrimes( p, k, func = getPreviousPrimeCandidateForAny ):
    if p < 1000000000 and getNthPrime( p ) < k:
        raise ValueError( 'There is no previous prime to 2.' )

    result = [ ]

    for _ in arange( 0, k ):
        p = getPreviousPrime( p, func )
        result.append( p )

    return result


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( 0 ) ] )
def getPreviousPrimesOperator( n, k ):
    return getPreviousPrimes( n, k, func=getPreviousPrimeCandidateForAny )


#******************************************************************************
#
#  getNthPrimeOperator
#
#******************************************************************************

def getNthPrime( arg ):
    n = int( arg )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 2

    if n == 2:
        return 3

    if n == 3:
        return 5

    if g.primeDataAvailable and n >= 1000000000:
        openPrimeCache( 'huge_primes' )

        maxIndex = g.cursors[ 'huge_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        currentIndex, p = g.cursors[ 'huge_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    elif g.primeDataAvailable and n >= 1000000:
        openPrimeCache( 'large_primes' )

        currentIndex, p = g.cursors[ 'large_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    elif g.primeDataAvailable and n >= 100:
        openPrimeCache( 'small_primes' )

        currentIndex, p = g.cursors[ 'small_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        currentIndex = 4
        p = 7

    while n > currentIndex:
        p = getNextPrime( p )
        currentIndex += 1

    return p


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthPrimeOperator( n ):
    return getNthPrime( n )


#******************************************************************************
#
#  findPrimeOperator
#
#******************************************************************************

def findPrime( arg ):
    target = int( arg )

    if target < 3:
        return 1, 2

    if target == 3:
        return 2, 3

    if target <= 5:
        return 3, 5

    if target <= 7:
        return 4, 7

    if target < 541:          # 100th prime
        currentIndex = 4
        p = 7
    elif g.primeDataAvailable and target <= 15485863:     # 1,000,000th prime
        openPrimeCache( 'small_primes' )

        currentIndex, p = g.cursors[ 'small_primes' ].execute(
            '''SELECT id, max( value ) FROM cache WHERE value <= ?''', ( target, ) ).fetchone( )
    elif g.primeDataAvailable and target <= 22801763489:  # 1,000,000,000th prime
        openPrimeCache( 'large_primes' )

        currentIndex, p = g.cursors[ 'large_primes' ].execute(
            '''SELECT id, max( value ) FROM cache WHERE value <= ?''', ( target, ) ).fetchone( )
    elif g.primeDataAvailable:
        openPrimeCache( 'huge_primes' )

        currentIndex, p = g.cursors[ 'huge_primes' ].execute(
            '''SELECT id, max( value ) FROM cache WHERE value <= ?''', ( target, ) ).fetchone( )
    else:
        currentIndex = 4
        p = 7

    while True:
        oldP = p
        p = getNextPrime( p )
        currentIndex += 1

        if p > target:
            return currentIndex - 1, oldP


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def findPrimeOperator( n ):
    return findPrime( n )[ 0 ]


#******************************************************************************
#
#  findQuadrupletPrimesOperator
#
#******************************************************************************

def findQuadrupletPrimes( arg ):
    n = int( arg )

    if n < 5:
        return 1, [ 5, 7, 11, 13 ]

    if n < 11:
        return 2, [ 11, 13, 17, 19 ]

    if g.primeDataAvailable:
        openPrimeCache( 'quad_primes' )

        currentIndex, p = g.cursors[ 'quad_primes' ].execute(
            '''SELECT id, max( value ) FROM cache WHERE value <= ?''', ( n, ) ).fetchone( )
    else:
        p = 11

    while True:
        # All quadruplet primes mod 30 == 11
        p += 30

        if isPrime( p ) and isPrime( p + 2 ) and isPrime( p + 6 ) and isPrime( p + 8 ):
            currentIndex += 1

            if p > n:
                return currentIndex, [ p, p + 2, p + 6, p + 8 ]


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def findQuadrupletPrimeOperator( n ):
    return findQuadrupletPrimes( n )[ 0 ]


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNextQuadrupletPrimeOperator( n ):
    return findQuadrupletPrimes( n )[ 1 ][ 0 ]


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNextQuadrupletPrimesOperator( n ):
    return findQuadrupletPrimes( n )[ 1 ]


#******************************************************************************
#
#  getNthQuadrupletPrimeOperator
#
#******************************************************************************

def getNthQuadrupletPrime( arg ):
    n = int( arg )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 5

    if n == 2:
        return 11

    if g.primeDataAvailable and n >= 10:
        openPrimeCache( 'quad_primes' )

        maxIndex = g.cursors[ 'quad_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        startingPlace, p = g.cursors[ 'quad_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        startingPlace = 2
        p = 11

    # after 5, the first of a prime quadruplet must be a number of the form 30n + 11
    while n > startingPlace:
        p += 30

        if isPrime( p ) and isPrime( p + 2 ) and isPrime( p + 6 ) and isPrime( p + 8 ):
            n -= 1

    return p


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthQuadrupletPrimeOperator( n ):
    return getNthQuadrupletPrime( n )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthQuadrupletPrimeListOperator( n ):
    p = getNthQuadrupletPrime( n )
    return [ p, fadd( p, 2 ), fadd( p, 6 ), fadd( p, 8 ) ]


#******************************************************************************
#
#  getNthIsolatedPrimeOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthIsolatedPrimeOperator( arg ):
    n = int( arg )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 2

    if n == 2:
        return 23

    if g.primeDataAvailable:
        openPrimeCache( 'isolated_primes' )

        currentIndex, p = g.cursors[ 'isolated_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        currentIndex = 2
        p = 23

    while n > currentIndex:
        p = getNextPrime( p )

        if not isPrime( p - 2 ) and not isPrime( p + 2 ):
            currentIndex += 1

    return p


#******************************************************************************
#
#  getNextTwinPrimeCandidate
#
#  Looking at ( p - 10 ) mod 30, the only twin prime candidates are 1, 7,
#  and 19.
#
#******************************************************************************

def getNextTwinPrimeCandidate( p ):
    f = ( p - 10 ) % 30

    if f == 1:
        p += 6
    elif f == 7:
        p += 12
    else:  # f == 19
        p += 12

    return p


#******************************************************************************
#
#  getNthTwinPrimeOperator
#
#******************************************************************************

def getNthTwinPrime( arg ):
    n = int( arg )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 3

    if n == 2:
        return 5

    if n == 3:
        return 11

    if g.primeDataAvailable and n >= 100:
        openPrimeCache( 'twin_primes' )

        maxIndex = g.cursors[ 'twin_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        currentIndex, p = g.cursors[ 'twin_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        currentIndex = 3
        p = 11

    while n > currentIndex:
        p = getNextPrime( p, getNextTwinPrimeCandidate )

        if isPrime( p + 2 ):
            currentIndex += 1

    return p


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthTwinPrimeOperator( n ):
    return getNthTwinPrime( n )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthTwinPrimeListOperator( n ):
    p = getNthTwinPrime( n )
    return [ p, fadd( p, 2 ) ]


#******************************************************************************
#
#  getNthBalancedPrimeOperator
#
#  returns the first balanced prime
#
#******************************************************************************

def getNthBalancedPrime( arg ):
    n = int( arg )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 5

    if g.primeDataAvailable:
        openPrimeCache( 'balanced_primes' )

        maxIndex = g.cursors[ 'balanced_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        currentIndex, p = g.cursors[ 'balanced_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )

        if n == currentIndex:
            return p

        # fill in the variables as if we'd been searching...
        secondPrevPrime = getPreviousPrime( p )
        prevPrime = p
        p = getNextPrime( prevPrime )
    else:
        currentIndex = 1
        p = 7
        prevPrime = 5
        secondPrevPrime = 3

    while n > currentIndex:
        secondPrevPrime = prevPrime
        prevPrime = p
        p = getNextPrime( p )

        if ( prevPrime - secondPrevPrime ) == ( p - prevPrime ):
            currentIndex += 1

    return prevPrime


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthBalancedPrimeOperator( n ):
    return getNthBalancedPrime( n )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthBalancedPrimeListOperator( n ):
    q = getNthBalancedPrime( n )
    p = getPreviousPrime( q )
    r = getNextPrime( q )

    return [ p, q, r ]


#******************************************************************************
#
#  getNthDoubleBalancedPrimeElement
#
#******************************************************************************

def getNthDoubleBalancedPrimeElement( arg, first = False ):
    n = int( arg )

    numberOfPrimes = 5
    center = numberOfPrimes // 2

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 18731

    if g.primeDataAvailable:
        openPrimeCache( 'double_balanced_primes' )

        maxIndex = g.cursors[ 'double_balanced_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )
        currentIndex, p = g.cursors[ 'double_balanced_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )

        if currentIndex == n:
            return p
    else:
        if n > 50:
            sys.stderr.write( 'The prime number cache data is not available.  This could take some time...\n' )

        # no cache... we'll do this the hard way
        currentIndex = 1
        p = 18731

    workingPrimes = [ 0 ] * numberOfPrimes
    workingPrimes[ center ] = p

    for i in range( 1, center + 1 ):
        workingPrimes[ center - i ] = getPreviousPrime( workingPrimes[ center - i + 1 ] )
        workingPrimes[ center + i ] = getNextPrime( workingPrimes[ center + i - 1 ] )

    while n > currentIndex:
        p = getNextPrime( workingPrimes[ -1 ] )
        workingPrimes.append( p )
        del workingPrimes[ 0 ]

        balanced = True

        for i in range( 1, center + 1 ):
            if ( workingPrimes[ center - i + 1 ] - workingPrimes[ center - i ] ) != \
               ( workingPrimes[ center + i ] - workingPrimes[ center + i - 1 ] ):
                balanced = False
                break

        if balanced:
            currentIndex += 1

    return workingPrimes[ 0 ] if first else workingPrimes[ center ]


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthDoubleBalancedPrimeListOperator( n ):
    p = getNthDoubleBalancedPrimeElement( n, first = False )

    result = [ 0, 0, p, 0, 0 ]
    middle = 2

    for i in range( 0, 2 ):
        result[ middle + i + 1 ] = getNextPrime( result[ middle + i ] )
        result[ middle - ( i + 1 ) ] = getPreviousPrime( result[ middle - i ] )

    return result


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthDoubleBalancedPrimeOperator( n ):
    return getNthDoubleBalancedPrimeElement( n )


#******************************************************************************
#
#  getNthTripleBalancedPrimeElement
#
#******************************************************************************

def getNthTripleBalancedPrimeElement( arg, first = False ):
    n = int( arg )

    numberOfPrimes = 7
    center = numberOfPrimes // 2

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 683783

    if g.primeDataAvailable:
        openPrimeCache( 'triple_balanced_primes' )

        maxIndex = g.cursors[ 'triple_balanced_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        currentIndex, p = g.cursors[ 'triple_balanced_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )

        if n == currentIndex:
            return p
    else:
        if n > 10:
            sys.stderr.write( 'The prime number cache data is not available.  This could take some time...\n' )

        # no cache... we'll do this the hard way
        currentIndex = 1
        p = 683783

    workingPrimes = [ 0 ] * numberOfPrimes
    workingPrimes[ center ] = p

    for i in range( 1, center + 1 ):
        workingPrimes[ center - i ] = getPreviousPrime( workingPrimes[ center - i + 1 ] )
        workingPrimes[ center + i ] = getNextPrime( workingPrimes[ center + i - 1 ] )

    while n > currentIndex:
        p = getNextPrime( workingPrimes[ -1 ] )
        workingPrimes.append( p )
        del workingPrimes[ 0 ]

        balanced = True

        for i in range( 1, center + 1 ):
            if ( workingPrimes[ center - i + 1 ] - workingPrimes[ center - i ] ) != \
               ( workingPrimes[ center + i ] - workingPrimes[ center + i - 1 ] ):
                balanced = False
                break

        if balanced:
            currentIndex += 1

    return workingPrimes[ 0 ] if first else workingPrimes[ center ]


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthTripleBalancedPrimeListOperator( n ):
    p = getNthTripleBalancedPrimeElement( n, first = True )

    result = [ 0, 0, 0, p, 0, 0, 0 ]
    middle = 3

    for i in range( 0, 3 ):
        result[ middle + i + 1 ] = getNextPrime( result[ middle + i ] )
        result[ middle - ( i + 1 ) ] = getPreviousPrime( result[ middle - i ] )

    return result


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthTripleBalancedPrimeOperator( n ):
    return getNthTripleBalancedPrimeElement( n )


#******************************************************************************
#
#  getNthQuadrupleBalancedPrimeElement
#
#******************************************************************************

def getNthQuadrupleBalancedPrimeElement( arg, first = False ):
    n = int( arg )

    numberOfPrimes = 9
    center = numberOfPrimes // 2

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 98303927

    if g.primeDataAvailable:
        openPrimeCache( 'quadruple_balanced_primes' )

        maxIndex = g.cursors[ 'quadruple_balanced_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        currentIndex, p = g.cursors[ 'quadruple_balanced_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )

        if n == currentIndex:
            return p
    else:
        if n > 10:
            sys.stderr.write( 'The prime number cache data is not available.  This could take some time...\n' )

        # no cache... we'll do this the hard way
        currentIndex = 1
        p = 98303927

    workingPrimes = [ 0 ] * numberOfPrimes
    workingPrimes[ center ] = p

    for i in range( 1, center + 1 ):
        workingPrimes[ center - i ] = getPreviousPrime( workingPrimes[ center - i + 1 ] )
        workingPrimes[ center + i ] = getNextPrime( workingPrimes[ center + i - 1 ] )

    while n > currentIndex:
        p = getNextPrime( workingPrimes[ -1 ] )
        workingPrimes.append( p )
        del workingPrimes[ 0 ]

        balanced = True

        for i in range( 1, center + 1 ):
            if ( workingPrimes[ center - i + 1 ] - workingPrimes[ center - i ] ) != \
               ( workingPrimes[ center + i ] - workingPrimes[ center + i - 1 ] ):
                balanced = False
                break

        if balanced:
            currentIndex += 1

    return workingPrimes[ 0 ] if first else workingPrimes[ center ]


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthQuadrupleBalancedPrimeListOperator( n ):
    p = getNthQuadrupleBalancedPrimeElement( n, first = True )

    result = [ 0, 0, 0, 0, p, 0, 0, 0, 0 ]
    middle = 4

    for i in range( 0, 4 ):
        result[ middle + i + 1 ] = getNextPrime( result[ middle + i ] )
        result[ middle - ( i + 1 ) ] = getPreviousPrime( result[ middle - i ] )

    return result


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthQuadrupleBalancedPrimeOperator( n ):
    return getNthQuadrupleBalancedPrimeElement( n )


#******************************************************************************
#
#  getNthSophiePrimeOperator
#
#******************************************************************************

def getNthSophiePrime( arg ):
    n = int( arg )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 2

    if n == 2:
        return 3

    if n == 3:
        return 5

    if g.primeDataAvailable and n >= 100:
        openPrimeCache( 'sophie_primes' )

        maxIndex = g.cursors[ 'sophie_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        currentIndex, p = g.cursors[ 'sophie_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        currentIndex = 4
        p = 11

    while n > currentIndex:
        p = getNextPrime( p )

        if isPrime( 2 * p + 1 ):
            currentIndex += 1

    return p


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthSophiePrimeOperator( n ):
    return getNthSophiePrime( n )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getSafePrimeOperator( n ):
    return fadd( fmul( getNthSophiePrime( n ), 2 ), 1 )


#******************************************************************************
#
#  getNthCousinPrimeOperator
#
#  http://oeis.org/A023200
#
#  Validate: rpn 1 52 range cousin 23200 oeis  -
#
#******************************************************************************

def getNthCousinPrime( arg ):
    n = int( arg )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 3

    if g.primeDataAvailable and n >= 100:
        openPrimeCache( 'cousin_primes' )

        maxIndex = g.cursors[ 'cousin_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        currentIndex, p = g.cursors[ 'cousin_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        currentIndex = 2
        p = 7

    while n > currentIndex:
        p = getNextPrime( p )

        if isPrime( p + 4 ):
            currentIndex += 1

    return p


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthCousinPrimeOperator( n ):
    return getNthCousinPrime( n )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthCousinPrimeListOperator( n ):
    p = getNthCousinPrime( n )
    return [ p, fadd( p, 4 ) ]


#******************************************************************************
#
#  getNextSexyPrimeCandidate
#
#  For a number ( p - 10 ) mod 30, the only sexy candidates are 1, 3, 7, 13,
#  21 and 27.
#
#******************************************************************************

def getNextSexyPrimeCandidate( p ):
    f = ( p - 10 ) % 30

    if f == 1:
        p += 2
    elif f == 3:
        p += 4
    elif f == 7:
        p += 6
    elif f == 13:
        p += 8
    elif f == 21:
        p += 6
    else:   # f == 27
        p += 4

    return p


#******************************************************************************
#
#  getNthSexyPrime
#
#******************************************************************************

def getNthSexyPrime( arg ):
    n = int( arg )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 5

    if g.primeDataAvailable and n >= 100:
        openPrimeCache( 'sexy_primes' )

        maxIndex = g.cursors[ 'sexy_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        startingPlace, p = g.cursors[ 'sexy_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        startingPlace = 2
        p = 7

    while n > startingPlace:
        p = getNextPrime( p, getNextSexyPrimeCandidate )

        if isPrime( p + 6 ):
            n -= 1

    return p


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthSexyPrimeOperator( n ):
    return getNthSexyPrime( n )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthSexyPrimeListOperator( n ):
    p = getNthSexyPrime( n )
    return [ p, fadd( p, 6 ) ]


#******************************************************************************
#
#  getNthSexyTriplet
#
#******************************************************************************

def getNthSexyTriplet( arg ):
    n = int( arg )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 7

    if g.primeDataAvailable and n >= 100:
        openPrimeCache( 'sexy_triplets' )

        maxIndex = g.cursors[ 'sexy_triplets' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        startingPlace, p = g.cursors[ 'sexy_triplets' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        startingPlace = 1
        p = 7

    f = p % 10

    while n > startingPlace:
        if f == 1:
            p += 6
            f = 7
        else:
            p += 4
            f = 1

        if isPrime( p ) and isPrime( p + 6 ) and isPrime( p + 12 ) and not isPrime( p + 18 ):
            n -= 1

    return p


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthSexyTripletOperator( n ):
    return getNthSexyTriplet( n )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthSexyTripletListOperator( n ):
    p = getNthSexyTriplet( n )
    return [ p, fadd( p, 6 ), fadd( p, 12 ) ]


#******************************************************************************
#
#  getNthSexyQuadruplet
#
#******************************************************************************

def getNthSexyQuadruplet( arg ):
    n = int( arg )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 5

    if n < 100:
        startingPlace = 2
        p = 11
    elif g.primeDataAvailable and n >= 100:
        openPrimeCache( 'sexy_quadruplets' )

        maxIndex = g.cursors[ 'sexy_quadruplets' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        startingPlace, p = g.cursors[ 'sexy_quadruplets' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )

    while n > startingPlace:
        p += 10

        if isPrime( p ) and isPrime( p + 6 ) and isPrime( p + 12 ) and isPrime( p + 18 ):
            n -= 1

    return p


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthSexyQuadrupletOperator( n ):
    return getNthSexyQuadruplet( n )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthSexyQuadrupletListOperator( n ):
    p = getNthSexyQuadruplet( n )
    return [ p, fadd( p, 6 ), fadd( p, 12 ), fadd( p, 18 ) ]


#******************************************************************************
#
#  getNextOctyPrimeCandidate
#
#  For a number ( p - 10 ) mod 30, the only sexy candidates are 1, 3, 9, 11,
#  13, 19, 21, 23 and 29.
#
#******************************************************************************

def getNextOctyPrimeCandidate( p ):
    f = ( p - 10 ) % 30

    if f == 1:
        p += 2
    elif f == 3:
        p += 6
    elif f == 9:
        p += 2
    elif f == 11:
        p += 2
    elif f == 13:
        p += 6
    elif f == 19:
        p += 2
    elif f == 21:
        p += 2
    elif f == 23:
        p += 6
    else:  # 29
        p += 2

    return p


#******************************************************************************
#
#  getNthOctyPrime
#
#******************************************************************************

def getNthOctyPrime( arg ):
    n = int( arg )

    if n < 1:
        raise ValueError( 'index must be > 0' )

    if n == 1:
        return 3

    if n == 2:
        return 5

    if g.primeDataAvailable:
        openPrimeCache( 'octy_primes' )

        maxIndex = g.cursors[ 'octy_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        startingPlace, p = g.cursors[ 'octy_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        startingPlace = 3
        p = 11

    while n > startingPlace:
        p = getNextPrime( p, getNextOctyPrimeCandidate )

        if isPrime( p + 8 ):
            n -= 1

    return p


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthOctyPrimeOperator( n ):
    return getNthOctyPrime( n )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthOctyPrimeListOperator( arg ):
    p = getNthOctyPrime( arg )
    return [ p, fadd( p, 8 ) ]


#******************************************************************************
#
#  getNthTripletPrimeList
#
#******************************************************************************

def getNthTripletPrimeList( arg ):
    n = int( arg )

    if n == 1:
        return [ 5, 7, 11 ]

    if n == 2:
        return [ 7, 11, 13 ]

    if n == 3:
        return [ 11, 13, 17 ]

    if g.primeDataAvailable and n >= 100:
        openPrimeCache( 'triplet_primes' )

        maxIndex = g.cursors[ 'triplet_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        currentIndex, p = g.cursors[ 'triplet_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        currentIndex = 3
        p = 11

    f = p % 10

    pPlus2 = False

    while n > currentIndex:
        if f == 1:
            p += 2
            f = 3
        elif f == 3:
            p += 4
            f = 7
        else:
            p += 4
            f = 1

        if isPrime( p ) and isPrime( p + 6 ):
            if isPrime( p + 2 ):
                pPlus2 = True
                currentIndex += 1
            elif isPrime( p + 4 ):
                pPlus2 = False
                currentIndex += 1

    return [ p, p + 2 if pPlus2 else p + 4, p + 6 ]


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthTripletPrimeListOperator( n ):
    return getNthTripletPrimeList( n )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthTripletPrimeOperator( arg ):
    return getNthTripletPrimeList( arg )[ 0 ]


#******************************************************************************
#
#  getNextQuintupletPrimeCandidate
#
#  For ( p - 10 ) mod 30, the only quintuplet prime candidates are:
#  1, 7, 21, or 27
#
#******************************************************************************

def getNextQuintupletPrimeCandidate( p ):
    f = p % 30

    if f == 7:
        p += 4
    else:
        p += 26

    return p


#******************************************************************************
#
#  getNthQuintupletPrimeOperator
#
#*****************************************************************************

def getNthQuintupletPrime( arg ):
    n = int( arg )

    if n == 1:
        return 5

    if n == 2:
        return 7

    if g.primeDataAvailable and n >= 10:
        openPrimeCache( 'quint_primes' )

        maxIndex = g.cursors[ 'quint_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        currentIndex, p = g.cursors[ 'quint_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        currentIndex = 3
        p = 11

    # after 5, the first of a prime quintuplet must be a number of the form 30n + 11
    while n > currentIndex:
        p = getNextPrime( p, getNextQuintupletPrimeCandidate )

        f = p % 30

        if ( ( f == 11 ) and isPrime( p + 2 ) and isPrime( p + 6 ) and
             isPrime( p + 8 ) and isPrime( p + 12 ) ):
            currentIndex += 1
        elif ( ( f == 7 ) and isPrime( p + 4 ) and isPrime( p + 6 ) and
               isPrime( p + 10 ) and isPrime( p + 12 ) ):
            currentIndex += 1

    return p


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthQuintupletPrimeOperator( n ):
    return getNthQuintupletPrime( n )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthQuintupletPrimeListOperator( n ):
    if n == 1:
        return [ 5, 7, 11, 13, 17 ]

    p = getNthQuintupletPrime( n )

    f = p % 10

    if f == 1:
        return [ p, fadd( p, 2 ), fadd( p, 6 ), fadd( p, 8 ), fadd( p, 12 ) ]

    if f == 7:
        return [ p, fadd( p, 4 ), fadd( p, 6 ), fadd( p, 10 ), fadd( p, 12 ) ]

    raise ValueError( 'internal error:  getNthQuintupletPrimeList is broken' )


#******************************************************************************
#
#  findQuintupletPrimes
#
#******************************************************************************

def findQuintupletPrimes( arg ):
    n = int( arg )

    if n < 5:
        return 1, [ 5, 7, 11, 13, 17 ]
    elif n < 7:
        return 2, [ 7, 11, 13, 17, 19 ]

    if g.primeDataAvailable:
        openPrimeCache( 'quint_primes' )

        currentIndex, p = g.cursors[ 'quint_primes' ].execute(
            '''SELECT id, max( value ) FROM cache WHERE value <= ?''', ( n, ) ).fetchone( )
    else:
        currentIndex = 3
        p = 11

    while True:
        p += 30

        f = p % 10

        if ( ( f == 1 ) and isPrime( p + 2 ) and isPrime( p + 6 ) and isPrime( p + 8 ) and isPrime( p + 12 ) ):
            currentIndex += 1

            if p > n:
                return currentIndex, [ p, fadd( p, 2 ), fadd( p, 6 ), fadd( p, 8 ), fadd( p, 12 ) ]
        elif ( ( f == 7 ) and isPrime( p + 4 ) and isPrime( p + 6 ) and isPrime( p + 10 ) and isPrime( p + 12 ) ):
            currentIndex += 1

            if p > n:
                return currentIndex, [ p, fadd( p, 4 ), fadd( p, 6 ), fadd( p, 10 ), fadd( p, 12 ) ]


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def findQuintupletPrimeOperator( n ):
    return findQuintupletPrimes( n )[ 0 ]


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNextQuintupletPrimeOperator( n ):
    return findQuintupletPrimes( n )[ 1 ][ 0 ]


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNextQuintupletPrimesOperator( n ):
    return findQuintupletPrimes( n )[ 1 ]


#******************************************************************************
#
#  findSextupletPrimes
#
#******************************************************************************

def findSextupletPrimes( arg ):
    n = int( arg )

    if n < 7:
        return 1, [ 7, 11, 13, 17, 19, 23 ]
    elif n < 97:
        return 2, [ 97, 101, 103, 107, 109, 113 ]

    if g.primeDataAvailable:
        openPrimeCache( 'sext_primes' )

        currentIndex, p = g.cursors[ 'sext_primes' ].execute(
            '''SELECT id, max( value ) FROM cache WHERE value <= ?''', ( n, ) ).fetchone( )
    else:
        currentIndex = 3
        p = 16057

    while True:
        p += 210

        if isPrime( p + 4 ) and isPrime( p + 6 ) and isPrime( p + 10 ) and \
           isPrime( p + 12 ) and isPrime( p + 16 ):
            currentIndex += 1

            if p > n:
                return currentIndex, [ p, fadd( p, 4 ), fadd( p, 6 ), fadd( p, 10 ), fadd( p, 12 ), fadd( p, 16 ) ]


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def findSextupletPrimeOperator( n ):
    return findSextupletPrimes( n )[ 0 ]


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNextSextupletPrimeOperator( n ):
    return findSextupletPrimes( n )[ 1 ][ 0 ]


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNextSextupletPrimesOperator( n ):
    return findSextupletPrimes( n )[ 1 ]


#******************************************************************************
#
#  findTripletPrimes
#
#******************************************************************************

def findTripletPrimes( arg ):
    n = int( arg )

    if n < 5:
        return 1, [ 5, 7, 11 ]

    if g.primeDataAvailable:
        openPrimeCache( 'triplet_primes' )

        currentIndex, p = g.cursors[ 'triplet_primes' ].execute(
            '''SELECT id, max( value ) FROM cache WHERE value <= ?''', ( n, ) ).fetchone( )
    else:
        currentIndex = 2
        p = 7

    while True:
        p = getNextPrime( p )

        if isPrime( p + 6 ):
            if isPrime( p + 2 ):
                currentIndex += 1

                if p > n:
                    return currentIndex, [ p, fadd( p, 2 ), fadd( p, 6 ) ]
            elif isPrime( p + 4 ):
                currentIndex += 1

                if p > n:
                    return currentIndex, [ p, fadd( p, 4 ), fadd( p, 6 ) ]


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def findTripletPrimeOperator( n ):
    return findTripletPrimes( n )[ 0 ]


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNextTripletPrimeOperator( n ):
    return findTripletPrimes( n )[ 1 ][ 0 ]


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNextTripletPrimesOperator( n ):
    return findTripletPrimes( n )[ 1 ]


#******************************************************************************
#
#  findTwinPrimes
#
#******************************************************************************

def findTwinPrimes( arg ):
    n = int( arg )

    if n < 5:
        return 1, [ 3, 5 ]

    if g.primeDataAvailable:
        openPrimeCache( 'twin_primes' )

        currentIndex, p = g.cursors[ 'twin_primes' ].execute(
            '''SELECT id, max( value ) FROM cache WHERE value <= ?''', ( n, ) ).fetchone( )
    else:
        currentIndex = 2
        p = 5

    while True:
        p = getNextPrime( p )

        if isPrime( p + 2 ):
            currentIndex += 1

            if p > n:
                return currentIndex, [ p, fadd( p, 2 ) ]


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def findTwinPrimeOperator( n ):
    return findTwinPrimes( n )[ 0 ]


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNextTwinPrimeOperator( n ):
    return findTwinPrimes( n )[ 1 ][ 0 ]


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNextTwinPrimesOperator( n ):
    return findTwinPrimes( n )[ 1 ]


#******************************************************************************
#
#  getNthSextupletPrimeOperator
#
#******************************************************************************

def getNthSextupletPrime( arg ):
    n = int( arg )

    if n == 1:
        return 7

    if n == 2:
        return 97

    if g.primeDataAvailable:
        openPrimeCache( 'sext_primes' )

        maxIndex = g.cursors[ 'sext_primes' ].execute(
            '''SELECT MAX( id ) FROM cache''' ).fetchone( )[ 0 ]

        if n > maxIndex:
            sys.stderr.write( '{:,} is above the max cached index of {:,}.  This could take some time...\n'.
                              format( n, maxIndex ) )

        startingPlace, p = g.cursors[ 'sext_primes' ].execute(
            '''SELECT MAX( id ), value FROM cache WHERE id <= ?''', ( int( n ), ) ).fetchone( )
    else:
        startingPlace = 2
        p = 97

    # all sets of prime sextuplets must start with 210x+97
    while n > startingPlace:
        p += 210

        if isPrime( p ) and isPrime( p + 4 ) and isPrime( p + 6 ) and \
           isPrime( p + 10 ) and isPrime( p + 12 ) + isPrime( p + 16 ):
            n -= 1

    return p


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthSextupletPrimeOperator( n ):
    return getNthSextupletPrime( n )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthSextupletPrimeListOperator( arg ):
    p = getNthSextupletPrime( arg )
    return [ p, fadd( p, 4 ), fadd( p, 6 ), fadd( p, 10 ), fadd( p, 12 ), fadd( p, 16 ) ]


#******************************************************************************
#
#  getNthPrimeRangeOperator
#
#******************************************************************************

def getNthPrimeRange( arg1, arg2 ):
    n = int( arg1 )
    count = int( arg2 )

    if count < 1:
        return [ ]

    # for primes below 7, we have to do it manually
    if n == 1:
        if count == 1:
            return [ 2 ]
        elif count == 2:
            return[ 2, 3 ]
        elif count == 3:
            return [ 2, 3, 5 ]
        else:
            result = [ 2, 3, 5, 7 ]
            count -= 3
            p = 7
    elif n == 2:
        if count == 1:
            return [ 3 ]
        elif count == 2:
            return [ 3, 5 ]
        else:
            result = [ 3, 5, 7 ]
            count -= 2
            p = 7
    elif n == 3:
        if count == 1:
            return [ 5 ]
        else:
            result = [ 5, 7 ]
            count -= 1
            p = 7
    else:
        p = getNthPrime( n )
        result = [ p ]

    found = 1

    while found < count:
        p = getNextPrimeCandidateForAny( p )

        if isPrime( p ):
            result.append( p )
            found += 1

    return result


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ), IntValidator( 0 ) ] )
def getNthPrimeRangeOperator( n, k ):
    return getNthPrimeRange( n, k )


#******************************************************************************
#
#  getNthSuperPrimeOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def getNthSuperPrimeOperator( arg ):
    return getNthPrime( getNthPrime( arg ) )


#******************************************************************************
#
#  getNthPolyPrimeOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ), IntValidator( 1 ) ] )
def getNthPolyPrimeOperator( n, poly ):
    result = getNthPrime( n )

    for _ in arange( 1, poly ):
        result = getNthPrime( result )

    return result


#******************************************************************************
#
#  getPrimesOperator
#
#******************************************************************************

def getPrimes( value, count ):
    for i in getNthPrimeRange( value, count ):
        yield i


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ), IntValidator( 1 ) ] )
def getPrimesOperator( n, k ):
    return RPNGenerator( getPrimes( n, k ) )


#******************************************************************************
#
#  getPrimeRangeOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ), IntValidator( 1 ) ] )
def getPrimeRangeOperator( start, end ):
    result = list( )

    for i in getNthPrimeRange( start, fadd( fsub( end, start ), 1 ) ):
        result.append( i )

    return result


#******************************************************************************
#
#  getNthPrimorialOperator
#
#******************************************************************************

def getNthPrimorial( n ):
    if n == 0:
        return 1

    return fprod( getPrimes( 1, n ) )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def getNthPrimorialOperator( n ):
    return getNthPrimorial( n )


#******************************************************************************
#
#  countCache
#
#******************************************************************************

def countCache( name ):
    openPrimeCache( name )
    return g.cursors[ name ].execute( '''SELECT count( id ) FROM cache''' ).fetchone( )[ 0 ]


#******************************************************************************
#
#  getMaxPrime
#
#******************************************************************************

def getMaxPrime( name ):
    openPrimeCache( name )
    return g.cursors[ name ].execute( '''SELECT max( id ), value FROM cache''' ).fetchone( )


#******************************************************************************
#
#  millerRabinPass
#
#  https://gist.github.com/sharnett/5479106
#
#******************************************************************************

def millerRabinPass( a, s, d, n ):
    aToPower = pow( a, d, n )

    if aToPower == 1:
        return True

    for _ in range( s - 1 ):
        if aToPower == n - 1:
            return True

        aToPower = ( aToPower * aToPower ) % n

    return aToPower == n - 1


#******************************************************************************
#
#  isPrimeMillerRabin
#
#  https://gist.github.com/sharnett/5479106
#  http://mathworld.wolfram.com/StrongPseudoprime.html
#
#  http://pi-e.de/Miller-Rabin-Pseudoprimzahlen.htm - check this out!
#
#******************************************************************************

def isPrimeMillerRabin( n ):
    if n < 2:
        return False

    if n < 2047:
        testPrimes = [ 2 ]
    elif n < 1373653:
        testPrimes = [ 2, 3 ]
    elif n < 9080191:
        testPrimes = [ 31, 73 ]
    elif n < 4759123141:
        testPrimes = [ 2, 7, 61 ]
    elif n < 2152302898747:
        testPrimes = [ 2, 3, 5, 7, 11 ]
    elif n < 3474749660383:
        testPrimes = [ 2, 3, 5, 7, 11, 13 ]
    elif n < 341550071728321:
        testPrimes = [ 2, 3, 5, 7, 11, 13, 17 ]
    elif n < 18446744073709551616:
        testPrimes = [ 2, 325, 9375, 28178, 450775, 9780504, 1795265022 ]
    elif n < 318665857834031151167461:
        testPrimes = [ 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37 ]
    elif n < 3317044064679887385961981:
        testPrimes = [ 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41 ]
    elif n < 6003094289670105800312596501 and g.zhangConjecturesAllowed:
        testPrimes = [ 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43 ]
    elif n < 59276361075595573263446330101 and g.zhangConjecturesAllowed:
        testPrimes = [ 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47 ]
    elif n < 564132928021909221014087501701 and g.zhangConjecturesAllowed:
        testPrimes = [ 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59 ]
    elif n < 1543267864443420616877677640751301 and g.zhangConjecturesAllowed:
        testPrimes = [ 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67 ]
    else:
        if g.zhangConjecturesAllowed:
            raise ValueError( str( n ) + ' is too large to use the deterministic Miller-Rabin test for 19 prime bases' )

        raise ValueError( str( n ) + ' is too large to use the deterministic Miller-Rabin test for 13 prime bases' )

    if n in testPrimes:
        return True

    d = n - 1
    s = 0

    while d % 2 == 0:
        d >>= 1
        s += 1

    for a in testPrimes:
        if not millerRabinPass( a, s, d, n ):
            return False

    return True


#******************************************************************************
#
#  isStrongPseudoprimeOperator
#
#******************************************************************************

def isStrongPseudoprime( n, k ):
    if n < k or fmod( n, 2 ) == 0 or isPrime( n ):
        return 0

    d = int( n - 1 )
    s = 0

    while d % 2 == 0:
        d >>= 1
        s += 1

    return 1 if millerRabinPass( int( k ), s, d, int( n ) ) else 0


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ), IntValidator( 1 ) ] )
def isStrongPseudoprimeOperator( n, k ):
    return isStrongPseudoprime( n, k )


#******************************************************************************
#
#  checkForPrimeData
#
#******************************************************************************

def checkForPrimeData( ):
    primeFile = Path( getUserDataPath( ) + os.sep + 'large_primes.cache' )

    if primeFile.is_file( ):
        g.primeDataAvailable = True
    else:
        g.primeDataAvailable = False



#!/usr/bin/env python

#******************************************************************************
#
#  rpnDateTime.py
#
#  rpnChilada date and time operations
#  copyright (c) 2021, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import calendar
import datetime

import arrow

from mpmath import mpf

from rpn.rpnDateTimeClass import RPNDateTime, getLocalTimeZone
from rpn.rpnGenerator import RPNGenerator
from rpn.rpnLocation import getTimeZoneName
from rpn.rpnMeasurementClass import RPNMeasurement
from rpn.rpnMeasurement import convertUnits
from rpn.rpnUtils import oneArgFunctionEvaluator, listArgFunctionEvaluator, twoArgFunctionEvaluator
from rpn.rpnValidator import argValidator, DateTimeValidator, IntValidator, ListValidator, YearValidator


#******************************************************************************
#
#  month and day constants
#
#******************************************************************************

MONDAY = 1
TUESDAY = 2
WEDNESDAY = 3
THURSDAY = 4
FRIDAY = 5
SATURDAY = 6
SUNDAY = 7

JANUARY = 1
FEBRUARY = 2
MARCH = 3
APRIL = 4
MAY = 5
JUNE = 6
JULY = 7
AUGUST = 8
SEPTEMBER = 9
OCTOBER = 10
NOVEMBER = 11
DECEMBER = 12


#******************************************************************************
#
#  convertToUnixTimeOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def convertToUnixTimeOperator( n ):
    try:
        result = RPNDateTime.parseDateTime( n.to( 'utc' ) ).int_timestamp
    except OverflowError:
        raise ValueError( 'out of range error' )
    except TypeError:
        raise ValueError( 'expected time value' )
    except OSError:
        raise ValueError( 'out of range error' )

    return result


#******************************************************************************
#
#  convertFromUnixTimeOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def convertFromUnixTimeOperator( n ):
    try:
        result = RPNDateTime.parseDateTime( int( n ) ).getLocalTime( )
    except OverflowError:
        raise ValueError( 'out of range error' )
    except TypeError:
        raise ValueError( 'expected time value' )
    except OSError:
        raise ValueError( 'out of range error' )

    return result


#******************************************************************************
#
#  convertToHMSOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def convertToHMSOperator( n ):
    return convertUnits( n, [ 'hour', 'minute', 'second' ] )


#******************************************************************************
#
#  convertToDHMSOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def convertToDHMSOperator( n ):
    return convertUnits( n, [ 'day', 'hour', 'minute', 'second' ] )


#******************************************************************************
#
#  convertToYDHMSOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def convertToYDHMSOperator( n ):
    return convertUnits( n, [ 'year', 'day', 'hour', 'minute', 'second' ] )


#******************************************************************************
#
#  makeJulianTimeOperator
#
#******************************************************************************

def makeJulianTime( n ):
    if isinstance( n, RPNGenerator ):
        return makeJulianTime( list( n ) )

    if len( n ) == 1:
        return RPNDateTime( n[ 0 ], 1, 1 )

    result = RPNDateTime( n[ 0 ], 1, 1 ).add( RPNMeasurement( n[ 1 ] - 1, 'day' ) )

    if len( n ) >= 3:
        result = result.replace( hour = n[ 2 ] )

    if len( n ) >= 4:
        result = result.replace( minute = n[ 3 ] )

    if len( n ) >= 5:
        result = result.replace( second = n[ 4 ] )

    if len( n ) >= 6:
        result = result.replace( microsecond = n[ 5 ] )

    return result


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def makeJulianTimeOperator( n ):
    return makeJulianTime( n )


#******************************************************************************
#
#  makeDateTimeOperator
#
#******************************************************************************

def makeDateTime( n ):
    if isinstance( n, ( RPNGenerator, int, mpf ) ):
        return makeDateTime( list( n ) )

    if isinstance( n[ 0 ], list ):
        return [ makeDateTime( i ) for i in n ]

    if len( n ) == 1:
        n.append( 1 )

    if len( n ) == 2:
        n.append( 1 )
    elif len( n ) > 7:
        n = n[ : 7 ]

    return RPNDateTime( *n )


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def makeDateTimeOperator( n ):
    return makeDateTime( n )


#******************************************************************************
#
#  getNowOperator
#
#******************************************************************************

def getNow( ):
    return RPNDateTime.now( tzinfo=getLocalTimeZone( ) )


def getNowOperator( ):
    return getNow( )


#******************************************************************************
#
#  getTodayOperator
#
#******************************************************************************

def getTodayOperator( ):
    now = getNow( )
    return RPNDateTime( *now.getYMD( ), dateOnly = True )


#******************************************************************************
#
#  getTomorrowOperator
#
#******************************************************************************

def getTomorrowOperator( ):
    now = getNow( )
    now = now + datetime.timedelta( days = 1 )
    return RPNDateTime( *now.getYMD( ), dateOnly = True )


#******************************************************************************
#
#  getYesterdayOperator
#
#******************************************************************************

def getYesterdayOperator( ):
    now = getNow( )
    now = now + datetime.timedelta( days = -1 )
    return RPNDateTime( *now.getYMD( ), dateOnly = True )


#******************************************************************************
#
#  calculateEasterOperator
#
#******************************************************************************

def calculateEaster( year ):
    '''This algorithm comes from Gauss.'''
    if isinstance( year, RPNDateTime ):
        year = year.year
    else:
        year = int( year )

    a = year % 19
    b = year // 100
    c = year % 100
    d = ( 19 * a + b - b // 4 - ( ( b - ( b + 8 ) // 25 + 1 ) // 3 ) + 15 ) % 30
    e = ( 32 + 2 * ( b % 4 ) + 2 * ( c // 4 ) - d - ( c % 4 ) ) % 7
    f = d + e - 7 * ( ( a + 11 * d + 22 * e ) // 451 ) + 114
    month = f // 31
    day = f % 31 + 1

    return RPNDateTime( year, month, day, dateOnly = True )


@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def calculateEasterOperator( year ):
    return calculateEaster( year )


#******************************************************************************
#
#  calculateAshWednesdayOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def calculateAshWednesdayOperator( year ):
    '''46 days before Easter (40 days, not counting Sundays)'''
    ashWednesday = calculateEaster( year ).add( RPNMeasurement( -46, 'day' ) )
    return RPNDateTime( *ashWednesday.getYMD( ), dateOnly = True )


#******************************************************************************
#
#  calculateGoodFridayOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def calculateGoodFridayOperator( year ):
    '''2 days before Easter'''
    goodFriday = calculateEaster( year ).add( RPNMeasurement( -2, 'day' ) )
    return RPNDateTime( *goodFriday.getYMD( ), dateOnly = True )


#******************************************************************************
#
#  getLastDayOfMonth
#
#******************************************************************************

def getLastDayOfMonth( year, month ):
    return calendar.monthrange( int( year ), int( month ) )[ 1 ]


#******************************************************************************
#
#  calculateNthWeekdayOfYear
#
#******************************************************************************

@argValidator( [ IntValidator( ), IntValidator( -53, 53 ), IntValidator( MONDAY, SUNDAY ) ] )
def calculateNthWeekdayOfYearOperator( year, nth, weekday ):
    ''' Monday = 1, etc., as per arrow, nth == -1 for last, etc.'''
    if isinstance( year, RPNDateTime ):
        year = year.year

    if nth > 0:
        firstDay = RPNDateTime( year, 1, 1 )

        firstWeekDay = weekday - firstDay.isoweekday( ) + 1

        if firstWeekDay < 1:
            firstWeekDay += 7

        result = RPNDateTime( year, 1, firstWeekDay ).add( RPNMeasurement( nth - 1, 'week' ) )
        result.setDateOnly( )

        return result

    if nth < 0:
        lastDay = RPNDateTime( year, 12, 31 )

        lastWeekDay = weekday - lastDay.isoweekday( )

        if lastWeekDay > 0:
            lastWeekDay -= 7

        lastWeekDay += 31

        result = RPNDateTime( year, 12, lastWeekDay, dateOnly = True ).add( RPNMeasurement( ( nth + 1 ), 'week' ) )
        result.setDateOnly( )

        return result

    raise ValueError( '0th weekday makes no sense in this context' )


#******************************************************************************
#
#  calculateNthWeekdayOfMonth
#
#******************************************************************************

def calculateNthWeekdayOfMonth( year, month, nth, weekday ):
    if weekday > SUNDAY or weekday < MONDAY:
        raise ValueError( 'day of week must be 1 - 7 (Monday to Sunday)' )

    if isinstance( year, RPNDateTime ):
        year = year.year

    firstDayOfWeek = arrow.Arrow( year, month, 1 ).isoweekday( )

    if nth < 0:
        day = ( ( weekday + 1 ) - firstDayOfWeek ) % 7

        while day <= getLastDayOfMonth( year, month ):
            day += 7

        day += nth * 7
    else:
        day = ( weekday - firstDayOfWeek + 1 ) + nth * 7

        if weekday >= firstDayOfWeek:
            day -= 7

    return RPNDateTime( year, month, day, dateOnly = True )


@argValidator( [ IntValidator( ), IntValidator( 1, 12 ), IntValidator( -5, 5 ), IntValidator( MONDAY, SUNDAY ) ] )
def calculateNthWeekdayOfMonthOperator( year, month, nth, weekday ):
    return calculateNthWeekdayOfMonth( year, month, nth, weekday )


#******************************************************************************
#
#  calculateThanksgivingOperator
#
#  the fourth Thursday in November
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def calculateThanksgivingOperator( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year

    return calculateNthWeekdayOfMonth( year, NOVEMBER, 4, THURSDAY )


#******************************************************************************
#
#  calculateLaborDayOperator
#
#  the first Monday in September
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def calculateLaborDayOperator( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year

    return calculateNthWeekdayOfMonth( year, SEPTEMBER, 1, MONDAY )


#******************************************************************************
#
#  calculateElectionDayOperator
#
#  the first Tuesday after the first Monday (so it's never on the 1st day)
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def calculateElectionDayOperator( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year

    result = calculateNthWeekdayOfMonth( year, NOVEMBER, 1, MONDAY )
    result.replace( day = result.day + 1 )

    return RPNDateTime( *result.getYMD( ), dateOnly = True )


#******************************************************************************
#
#  calculateMemorialDayOperator
#
#  the last Monday in May (4th or 5th Monday)
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def calculateMemorialDayOperator( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year

    return calculateNthWeekdayOfMonth( year, MAY, -1, MONDAY )


#******************************************************************************
#
#  calculatePresidentsDayOperator
#
#  the third Monday in February
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def calculatePresidentsDayOperator( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year

    return calculateNthWeekdayOfMonth( year, FEBRUARY, 3, MONDAY )


#******************************************************************************
#
#  calculateMartinLutherKingDayOperator
#
#  the third Monday in January
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def calculateMartinLutherKingDayOperator( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year

    return calculateNthWeekdayOfMonth( year, JANUARY, 3, MONDAY )


#******************************************************************************
#
#  calculateColumbusDayOperator
#
#  the second Monday in October
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def calculateColumbusDayOperator( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year

    return calculateNthWeekdayOfMonth( year, OCTOBER, 2, MONDAY )


#******************************************************************************
#
#  calculateMothersDayOperator
#
#  the second Sunday in May (in the U.S. and most other countries)
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def calculateMothersDayOperator( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year

    return calculateNthWeekdayOfMonth( year, MAY, 2, SUNDAY )


#******************************************************************************
#
#  calculateFathersDayOperator
#
#  the third Sunday in June (in the U.S. and most other countries)
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def calculateFathersDayOperator( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year

    return calculateNthWeekdayOfMonth( year, JUNE, 3, SUNDAY )


#******************************************************************************
#
#  getNewYearsDayOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def getNewYearsDayOperator( year ):
    return RPNDateTime( year, JANUARY, 1, dateOnly = True )


#******************************************************************************
#
#  getVeteransDayOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def getVeteransDayOperator( year ):
    return RPNDateTime( year, NOVEMBER, 11, dateOnly = True )


#******************************************************************************
#
#  getIndependenceDayOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def getIndependenceDayOperator( year ):
    return RPNDateTime( year, JULY, 4, dateOnly = True )


#******************************************************************************
#
#  getChristmasDayOperator
#
#******************************************************************************

def getChristmasDay( year ):
    return RPNDateTime( year, DECEMBER, 25, dateOnly = True )


@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def getChristmasDayOperator( year ):
    return getChristmasDay( year )


#******************************************************************************
#
#  calculateAdventOperator
#
#  4 Sundays before Christmas
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def calculateAdventOperator( year ):
    firstAdvent = getChristmasDay( year ).add( RPNMeasurement( -3, 'week' ) )
    firstAdvent = firstAdvent.subtract( RPNMeasurement( getWeekday( firstAdvent ), 'day'  ) )

    return RPNDateTime( *firstAdvent.getYMD( ), dateOnly = True )


#******************************************************************************
#
#  getEpiphanyDayOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def getEpiphanyDayOperator( year ):
    return RPNDateTime( year, 1, 6, dateOnly = True )


#******************************************************************************
#
#  calculatePentecostSundayOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def calculatePentecostSundayOperator( year ):
    return RPNDateTime( *( calculateEaster( year ).add( RPNMeasurement( 7, 'weeks' ) ).getYMD( ) ),
                        dateOnly = True )


#******************************************************************************
#
#  calculateAscensionThursdayOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def calculateAscensionThursdayOperator( year ):
    '''
I don't know why Ascension is 39 days after Easter instead of 40, but that's
how the math works out.  It's the 40th day of the Easter season.

Or as John Wright says, "Catholics can't count."  I think it stems from the
Church being created before the number 0.
    '''
    return RPNDateTime( *calculateEaster( year ).add( RPNMeasurement( 39, 'days' ) ).getYMD( ),
                        dateOnly = True )


#******************************************************************************
#
#  calculateDSTStartOperator
#
#  the second Sunday in March
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def calculateDSTStartOperator( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year

    if year >= 2007:
        return calculateNthWeekdayOfMonth( year, MARCH, 2, SUNDAY )

    if year == 1974:
        return RPNDateTime( 1974, JANUARY, 7, dateOnly = True )

    if year >= 1967:
        return calculateNthWeekdayOfMonth( year, APRIL, 1, SUNDAY )

    raise ValueError( 'DST was not standardized before 1967' )


#******************************************************************************
#
#  calculateDSTEndOperator
#
#  the first Sunday in November
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def calculateDSTEndOperator( year ):
    if isinstance( year, RPNDateTime ):
        year = year.year

    if year >= 2007:
        return calculateNthWeekdayOfMonth( year, NOVEMBER, 1, SUNDAY )

    if year == 1974:
        return RPNDateTime( 1974, DECEMBER, 31, dateOnly = True )  # technically DST never ended in 1974

    if year >= 1967:
        return calculateNthWeekdayOfMonth( year, OCTOBER, -1, SUNDAY )

    raise ValueError( 'DST was not standardized before 1967' )


#******************************************************************************
#
#  getWeekdayOperator
#
#******************************************************************************

def getWeekday( n ):
    return n.weekday( ) + 1


@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getWeekdayOperator( n ):
    return getWeekday( n )


#******************************************************************************
#
#  getWeekdayNameOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getWeekdayNameOperator( n ):
    return calendar.day_name[ getWeekday( n ) - 1 ]


#******************************************************************************
#
#  getYearOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getYearOperator( n ):
    return n.year


#******************************************************************************
#
#  getMonthOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getMonthOperator( n ):
    return n.month


#******************************************************************************
#
#  getDayOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getDayOperator( n ):
    return n.day


#******************************************************************************
#
#  getHourOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getHourOperator( n ):
    return n.hour


#******************************************************************************
#
#  getMinuteOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getMinuteOperator( n ):
    return n.minute


#******************************************************************************
#
#  getSecondOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getSecondOperator( n ):
    return n.second


#******************************************************************************
#
#  isDST
#
#******************************************************************************

def isDST( dateTime, timeZone ):
    return dateTime.astimezone( timeZone ).dst( ) != datetime.timedelta( 0 )


#******************************************************************************
#
#  setUTCOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getUTCOperator( dt ):
    return dt.to( 'utc' )


#******************************************************************************
#
#  setLocalTimeOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getLocalTimeOperator( dt ):
    return dt.getLocalTime( )


#******************************************************************************
#
#  setTimeZoneOperator
#
#******************************************************************************

def setTimeZone( dt, timezone ):
    try:
        tz = arrow.now( timezone ).tzinfo
    except TypeError:
        tz = arrow.now( getTimeZoneName( timezone ) ).tzinfo
    except arrow.parser.ParserError:
        tz = arrow.now( getTimeZoneName( timezone ) ).tzinfo
    #except Exception as e:
    #    import sys
    #    print('Whew!', sys.exc_info()[0], 'occurred.')

    dt = RPNDateTime.parseDateTime( dt )
    dt = dt.replace( tzinfo=tz )
    return dt


@twoArgFunctionEvaluator( )
#@argValidator( [ DateTimeValidator( ) ] )
def setTimeZoneOperator( dt, timezone ):
    return setTimeZone( dt, timezone )


#******************************************************************************
#
#  convertTimeZoneOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
#@argValidator( [ DateTimeValidator( ) ] )
def convertTimeZoneOperator( dt, timezone ):
    try:
        tz = arrow.now( timezone ).tzinfo
    except arrow.parser.ParserError:
        tz = arrow.now( getTimeZoneName( timezone ) ).tzinfo

    return dt.to( tz )



#!/usr/bin/env python

#******************************************************************************
#
#  rpnLocationClass.py
#
#  rpnChilada location class declaration
#  copyright (c) 2021, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import ephem

from mpmath import fdiv, fmul, mpmathify, pi
from skyfield.api import Topos


#******************************************************************************
#
#  class RPNLocation
#
#  The observer class measures lat/long in radians, but no one else does,
#  so the methods assume degrees.
#
#******************************************************************************

class RPNLocation( ):
    '''This class represents a location on the surface of the Earth.'''
    observer = None
    name = None

    def __init__( self, *_, **kwargs ):
        self.name = kwargs.get( 'name', None )
        observer = kwargs.get( 'observer', None )
        lat = kwargs.get( 'lat', None )
        long = kwargs.get( 'long', None )

        if observer:
            self.observer = observer
        else:
            self.observer = ephem.Observer( )

        if lat:
            self.setLat( lat )

        if long:
            self.setLong( long )

    def setObserver( self, observer ):
        self.name = ''
        self.observer.lat = observer.lat
        self.observer.long = observer.long
        self.observer.epoch = observer.epoch
        self.observer.date = observer.date
        self.observer.elevation = observer.elevation
        self.observer.temp = observer.temp
        self.observer.pressure = observer.pressure

    def getName( self ):
        return self.name

    # latitude in degrees
    def getLat( self ):
        return fdiv( fmul( mpmathify( float( self.observer.lat ) ), 180 ), pi )

    # longitude in degrees
    def getLong( self ):
        return fdiv( fmul( mpmathify( float( self.observer.long ) ), 180 ), pi )

    def getDate( self ):
        return self.observer.date

    def getEpoch( self ):
        return self.observer.epoch

    def getElevation( self ):
        return self.observer.elevation

    def getTemp( self ):
        return self.observer.temp

    def getPressure( self ):
        return self.observer.pressure

    def setName( self, value ):
        self.name = value

    # set latitude in degrees
    def setLat( self, value ):
        self.observer.lat = fmul( fdiv( value, 180 ), pi )

    # set longitude in degrees
    def setLong( self, value ):
        self.observer.long = fmul( fdiv( value, 180 ), pi )

    def setDate( self, value ):
        self.observer.date = value

    def setEpoch( self, value ):
        self.observer.epoch = value

    def setElevation( self, value ):
        self.observer.elevation = value

    def setTemp( self, value ):
        self.observer.temp = value

    def setPressure( self, value ):
        self.observer.pressure = value

    def getTopos( self ):
        return Topos( latitude_degrees=float( self.getLat( ) ),
                      longitude_degrees=float( self.getLong( ) ),
                      elevation_m=self.getElevation( ) )



#!/usr/bin/env python

#******************************************************************************
#
#  rpnBase.py
#
#  rpnChilada base conversion functions
#  copyright (c) 2021, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

from mpmath import chop, fac, fac2, fadd, fdiv, floor, fmul, fsub, log, mp, nint, phi, power

from rpn.rpnNumberTheory import getNthLucasNumber
from rpn.rpnPolytope import getNthPolygonalNumber
from rpn.rpnPrimeUtils import getNthPrimorial

import rpn.rpnGlobals as g


#******************************************************************************
#
#  specialBaseFunctons
#
#******************************************************************************

specialBaseFunctions = {
    g.facBase         : fac,
    g.doublefacBase   : fac2,
    g.squareBase      : lambda n: power( n, 2 ),
    g.lucasBase       : getNthLucasNumber,
    g.triangularBase  : lambda n: getNthPolygonalNumber( n, 3 ),
    g.primorialBase   : lambda n: getNthPrimorial( n - 1 ),
}


#******************************************************************************
#
#  specialBaseNames
#
#******************************************************************************

specialBaseNames = {
    'phi'               : g.phiBase,
    'fib'               : g.fibBase,
    'fibonacci'         : g.fibBase,
    'fac'               : g.facBase,
    'factorial'         : g.facBase,
    '!'                 : g.facBase,
    'factorial2'        : g.doublefacBase,
    'double_factorial'  : g.doublefacBase,
    'fac2'              : g.doublefacBase,
    'double_fac'        : g.doublefacBase,
    '!!'                : g.doublefacBase,
    'square'            : g.squareBase,
    'sqr'               : g.squareBase,
    'lucas'             : g.lucasBase,
    'triangular'        : g.triangularBase,
    'tri'               : g.triangularBase,
    'primorial'         : g.primorialBase,
    'e'                 : g.eBase,
    'pi'                : g.piBase,
    'root2'             : g.sqrt2Base,
    'sqrt2'             : g.sqrt2Base,
}


#******************************************************************************
#
#  convertToSpecialBase
#
#******************************************************************************

def convertToSpecialBase( value, baseFunction, outputBaseDigits = False, numerals = g.defaultNumerals ):
    '''
    This version supports arbitrary non-constant bases.  The place value is
    determined by the function passed in.  The function takes a single argument
    which represents the place, and it returns the value that that place
    represents.  As an example for base 10, the function would return 10^n for
    argument n.
    '''
    if value == 0:
        return 0

    if value < 0:
        raise ValueError( 'Negative numbers are not supported.' )

    if outputBaseDigits:
        result = [ ]
    else:
        result = ''

    positionValues = [ ]

    position = 1
    positionValue = baseFunction( position )

    while positionValue <= value:
        positionValues.append( positionValue )

        position += 1
        positionValue = baseFunction( position )

    if outputBaseDigits:
        result = [ ]
    else:
        result = ''

    remaining = value

    while positionValues:
        base = positionValues.pop( )

        digit = floor( fdiv( remaining, base ) )

        if outputBaseDigits:
            result.append( digit )
        else:
            result += numerals[ int( digit ) ]

        remaining = fsub( remaining, fmul( digit, base ) )

    return result


#******************************************************************************
#
#  convertToIterativeBase
#
#******************************************************************************

#def convertToIterativeBase( value, baseFunction, outputBaseDigits, numerals ):
#    return 'fred'


#******************************************************************************
#
#  convertToNonintegerBase
#
#******************************************************************************

def convertToNonintegerBase( num, base ):
    epsilon = power( 10, -( mp.dps - 3 ) )
    minPlace = -( floor( mp.dps / log( base ) ) )

    output = ''
    integer = ''

    remaining = num

    # find starting place
    place = int( floor( log( remaining, base ) ) )

    while remaining > epsilon:
        if place < minPlace:
            break

        if place == -1:
            integer = output
            output = ''

        placeValue = power( base, place )

        value = fdiv( remaining, placeValue )

        value = fmul( value, power( 10, mp.dps - 3 ) )
        value = nint( value )
        value = fdiv( value, power( 10, mp.dps - 3 ) )

        value = floor( value )
        remaining = chop( fsub( remaining, fmul( placeValue, value ) ) )

        output += str( value )[ : -2 ]

        place -= 1

    if place >= 0:
        integer = output + '0' * ( place + 1 )
        output = ''

    if integer == '':
        return output, ''

    return integer, output


#******************************************************************************
#
#  convertToPhiBase
#
#******************************************************************************

def convertToPhiBase( num ):
    epsilon = power( 10, -( mp.dps - 3 ) )

    output = ''
    integer = ''

    start = True
    previousPlace = 0
    remaining = num

    originalPlace = 0

    while remaining > epsilon:
        place = int( floor( log( remaining, phi ) ) )

        if start:
            output = '1'
            start = False
            originalPlace = place
        else:
            if place < -( originalPlace + 1 ):
                break

            for i in range( previousPlace, place + 1, -1 ):
                output += '0'

                if i == 1:
                    integer = output
                    output = ''

            output += '1'

            if place == 0:
                integer = output
                output = ''

        previousPlace = place
        remaining -= power( phi, place )

    if integer == '':
        return output, ''

    return integer, output


#******************************************************************************
#
#  convertToFibBase
#
#******************************************************************************

def convertToFibBase( value ):
    '''
    Returns a string with Fibonacci encoding for n (n >= 1).

    adapted from https://en.wikipedia.org/wiki/Fibonacci_coding
    '''

    result = ''

    n = value

    if n >= 1:
        a = 1
        b = 1

        c = fadd( a, b )    # next Fibonacci number
        fibs = [ b ]        # list of Fibonacci numbers, starting with F(2), each <= n

        while n >= c:
            fibs.append( c )  # add next Fibonacci number to end of list
            a = b
            b = c
            c = fadd( a, b )

        for fibnum in reversed( fibs ):
            if n >= fibnum:
                n = fsub( n, fibnum )
                result = result + '1'
            else:
                result = result + '0'

    return result



#!/usr/bin/env python

#******************************************************************************
#
#  rpnMeasurement.py
#
#  rpnChilada measurement operators and unit conversion
#  copyright (c) 2021, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

from mpmath import fdiv, fmul, nstr, mpmathify

from rpn.rpnDebug import debugPrint
from rpn.rpnGenerator import RPNGenerator
from rpn.rpnMeasurementClass import RPNMeasurement
from rpn.rpnUnitClasses import RPNUnits
from rpn.rpnUtils import oneArgFunctionEvaluator
from rpn.rpnValidator import argValidator, MeasurementValidator, MultiplicativeValidator

import rpn.rpnGlobals as g


#******************************************************************************
#
#  convertUnits
#
#******************************************************************************

def convertUnits( unit1, unit2 ):
    if isinstance( unit1, RPNGenerator ):
        unit1 = list( unit1 )

        if len( unit1 ) == 1:
            unit1 = unit1[ 0 ]

    if isinstance( unit2, RPNGenerator ):
        unit2 = list( unit2 )

        if len( unit2 ) == 1:
            unit2 = unit2[ 0 ]

    if isinstance( unit1, list ):
        result = [ ]

        for unit in unit1:
            result.append( convertUnits( unit, unit2 ) )

        return result

    if not isinstance( unit1, RPNMeasurement ):
        raise ValueError( 'cannot convert non-measurements' )

    if isinstance( unit2, list ):
        return unit1.convertValue( unit2 )

    if isinstance( unit2, ( str, RPNUnits ) ):
        measurement = RPNMeasurement( 1, unit2 )
        return RPNMeasurement( unit1.convertValue( measurement ), unit2 )

    if not isinstance( unit2, ( list, str, RPNUnits, RPNMeasurement ) ):
        raise ValueError( 'cannot convert non-measurements' )

    debugPrint( 'convertUnits' )
    debugPrint( 'unit1:', unit1.getUnitTypes( ) )
    debugPrint( 'unit2:', unit2.getUnitTypes( ) )

    newValue = unit1.convertValue( unit2 )

    debugPrint( '*** value:', newValue )

    return RPNMeasurement( newValue, unit2.units )


#@twoArgFunctionEvaluator( )  # This breaks the multiple unit conversion!
@argValidator( [ MeasurementValidator( ), MeasurementValidator( ) ] )
def convertUnitsOperator( unit1, unit2 ):
    return convertUnits( unit1, unit2 )


#******************************************************************************
#
#  convertToDMSOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MeasurementValidator( ) ] )
def convertToDMSOperator( n ):
    return convertUnits( n, [ 'degree', 'arcminute', 'arcsecond' ] )


#******************************************************************************
#
#  estimateOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MeasurementValidator( ) ] )
def estimateOperator( measurement ):
    if not isinstance( measurement, RPNMeasurement ):
        raise TypeError( 'incompatible type for estimating' )

    unitType = None

    dimensions = measurement.getDimensions( )

    for key, basicUnitType in g.basicUnitTypes.items( ):
        if dimensions == basicUnitType.dimensions:
            unitType = key
            break

    if unitType is None:
        return 'No estimates are available for this unit type'

    unitTypeInfo = g.basicUnitTypes[ unitType ]

    unit = RPNMeasurement( 1, unitTypeInfo.baseUnit )
    value = RPNMeasurement( measurement.convertValue( unit ), unit.units ).value

    if not unitTypeInfo.estimateTable:
        return 'No estimates are available for this unit type (' + unitType + ').'

    matchingKeys = [ key for key in unitTypeInfo.estimateTable if key <= value ]

    if matchingKeys:
        estimateKey = max( matchingKeys )

        multiple = fdiv( value, estimateKey )

        return 'approximately ' + nstr( multiple, 3 ) + ' times ' + \
               unitTypeInfo.estimateTable[ estimateKey ]
    else:
        estimateKey = min( key for key in unitTypeInfo.estimateTable )

        multiple = fdiv( estimateKey, value )

        return 'approximately ' + nstr( multiple, 3 ) + ' times smaller than ' + \
               unitTypeInfo.estimateTable[ estimateKey ]


#******************************************************************************
#
#  applyNumberValueToUnit
#
#  We have to treat constant units differently because they can become plain
#  numbers.
#
#******************************************************************************

def applyNumberValueToUnit( number, term, constant ):
    if isinstance( term, RPNUnits ):
        value = RPNMeasurement( number, term )
        value = value.normalizeUnits( )

        if value.units == RPNUnits( '_null_unit' ):
            value = mpmathify( 1 )
    elif constant:
        if g.constantOperators[ term ].unit:
            value = RPNMeasurement( fmul( number, mpmathify( g.constantOperators[ term ].value ) ),
                                    g.constantOperators[ term ].unit )
        else:
            value = fmul( number, g.constantOperators[ term ].value )
    else:
        if g.unitOperators[ term ].unitType == 'constant':
            value = RPNMeasurement( number, term ).convertValue( RPNMeasurement( 1, { 'unity' : 1 } ) )
        else:
            value = RPNMeasurement( number, term )

    return value


#******************************************************************************
#
#  getDimensions
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getDimensionsOperator( n ):
    if isinstance( n, RPNMeasurement ):
        return n.getDimensions( )
    else:
        return n


#******************************************************************************
#
#  convertToBaseUnitsOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MeasurementValidator( ) ] )
def convertToBaseUnitsOperator( n ):
    if isinstance( n, RPNMeasurement ):
        return n.convertToPrimitiveUnits( )
    else:
        return n


#******************************************************************************
#
#  convertToPrimitiveUnitsOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MeasurementValidator( ) ] )
def convertToPrimitiveUnitsOperator( n ):
    if isinstance( n, RPNMeasurement ):
        return n.convertToPrimitiveUnits( )
    else:
        return n


#******************************************************************************
#
#  invertUnitsOperator
#
#  invert the units and take the reciprocal of the value to create an
#  equivalent measurement
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MeasurementValidator( ) ] )
def invertUnitsOperator( measurement ):
    if not isinstance( measurement, RPNMeasurement ):
        raise ValueError( 'cannot invert non-measurements' )

    return measurement.getInverted( )



#!/usr/bin/env python

#******************************************************************************
#
#  rpnGeometry.py
#
#  rpnChilada geometry operators
#  copyright (c) 2021, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

from mpmath import cos, cot, fadd, fdiv, fmul, fprod, fsub, gamma, power, pi, \
                   root, sin, sqrt, tan

from rpn.rpnList import getProduct, getSum
from rpn.rpnMath import add, divide, getPower, getRoot, multiply, subtract
from rpn.rpnMeasurementClass import RPNMeasurement
from rpn.rpnUtils import oneArgFunctionEvaluator, twoArgFunctionEvaluator
from rpn.rpnValidator import argValidator, IntValidator, LengthValidator, MeasurementValidator


#******************************************************************************
#
#  getRegularPolygonArea
#
#  based on having sides of edge length k
#
#  http://www.mathopenref.com/polygonregulararea.html
#
#******************************************************************************

def getRegularPolygonArea( n, k ):
    return multiply( fdiv( n, fmul( 4, tan( fdiv( pi, n ) ) ) ), getPower( k, 2 ) ).convert( 'meter^2' )


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 3 ), LengthValidator( ) ] )
def getRegularPolygonAreaOperator( n, k ):
    return getRegularPolygonArea( n, k )


#******************************************************************************
#
#  getKSphereRadiusOperator
#
#  n - measurement
#  k - dimension
#
#  n needs to be an RPNMeasurement so getKSphereRadius can tell if it's an
#  area or a volume and use the correct formula.
#
#******************************************************************************

def getKSphereRadius( n, k ):
    if k < 3:
        raise ValueError( 'the number of dimensions must be at least 3' )

    if not isinstance( n, RPNMeasurement ):
        return RPNMeasurement( n, 'meter' )

    dimensions = n.getDimensions( )

    if dimensions == { 'length' : 1 }:
        return n

    if dimensions == { 'length' : int( k - 1 ) }:
        area = n.convertValue( RPNMeasurement( 1, [ { 'meter' : int( k - 1 ) } ] ) )

        result = root( fdiv( fmul( area, gamma( fdiv( k, 2 ) ) ),
                             fmul( 2, power( pi, fdiv( k, 2 ) ) ) ), fsub( k, 1 ) )

        return RPNMeasurement( result, [ { 'meter' : 1 } ] )

    if dimensions == { 'length' : int( k ) }:
        volume = n.convertValue( RPNMeasurement( 1, [ { 'meter' : int( k ) } ] ) )

        result = root( fmul( fdiv( gamma( fadd( fdiv( k, 2 ), 1 ) ),
                                   power( pi, fdiv( k, 2 ) ) ),
                             volume ), k )

        return RPNMeasurement( result, [ { 'meter' : 1 } ] )

    raise ValueError( 'incompatible measurement type for computing the radius: ' + str( dimensions ) )


@twoArgFunctionEvaluator( )
@argValidator( [ MeasurementValidator( 0 ), IntValidator( 3 ) ] )
def getKSphereRadiusOperator( n, k ):
    return getKSphereRadius( n, k )


@oneArgFunctionEvaluator( )
@argValidator( [ MeasurementValidator( 0 ) ] )
def getSphereRadiusOperator( n ):
    return getKSphereRadius( n, 3 )


#******************************************************************************
#
#  getKSphereSurfaceAreaOperator
#
#  https://en.wikipedia.org/wiki/N-sphere#Volume_and_surface_area
#
#  n - measurement
#  k - dimension
#
#  If n is a length, then it is taken to be the radius.  If it is a volume
#  then it is taken to be the volume.  If it is an area, then it is returned
#  unchanged.  Other measurement types cause an exception.
#
#******************************************************************************

def getKSphereSurfaceArea( n, k ):
    if not isinstance( n, RPNMeasurement ):
        return getKSphereSurfaceArea( n, RPNMeasurement( n, 'meter' ) )

    dimensions = n.getDimensions( )

    if dimensions == { 'length' : 1 }:
        m = n.convertValue( RPNMeasurement( 1, [ { 'meter' : 1 } ] ) )

        result = fmul( fdiv( fmul( power( pi, fdiv( k, 2 ) ), 2 ), gamma( fdiv( k, 2 ) ) ), power( m, fsub( k, 1 ) ) )

        return RPNMeasurement( result, [ { 'meter' : int( k - 1 ) } ] )

    if dimensions == { 'length' : int( k - 1 ) }:
        return n

    if dimensions == { 'length' : int( k ) }:
        radius = getKSphereRadius( n, k )
        return getKSphereSurfaceArea( radius, k )

    raise ValueError( 'incompatible measurement type for computing the surface area' )


@twoArgFunctionEvaluator( )
@argValidator( [ MeasurementValidator( 0 ), IntValidator( 3 ) ] )
def getKSphereSurfaceAreaOperator( n, k ):
    return getKSphereSurfaceArea( n, k )


@oneArgFunctionEvaluator( )
@argValidator( [ MeasurementValidator( 0 ) ] )
def getSphereAreaOperator( n ):
    return getKSphereSurfaceArea( n, 3 )


#******************************************************************************
#
#  getKSphereVolumeOperator
#
#  https://en.wikipedia.org/wiki/N-sphere#Volume_and_surface_area
#
#  n - measurement
#  k - dimension
#
#  If n is a length, then it is taken to be the radius.  If it is an area
#  then it is taken to be the surface area.  If it is a volume, then it is
#  returned unchanged.  Other measurement types cause an exception.
#
#******************************************************************************

def getKSphereVolume( n, k ):
    if not isinstance( n, RPNMeasurement ):
        return getKSphereVolume( RPNMeasurement( n, 'meter' ), k )

    dimensions = n.getDimensions( )
    m = n.value

    if dimensions == { 'length' : 1 }:
        m = n.convertValue( RPNMeasurement( 1, [ { 'meter' : 1 } ] ) )

        result = fmul( fdiv( power( pi, fdiv( k, 2 ) ), gamma( fadd( fdiv( k, 2 ), 1 ) ) ), power( m, k ) )

        return RPNMeasurement( result, [ { 'meter' : k } ] )

    if dimensions == { 'length' : int( k - 1 ) }:
        radius = getKSphereRadius( n, k )
        return getKSphereVolume( radius, k )

    if dimensions == { 'length' : int( k ) }:
        return n

    raise ValueError( 'incompatible measurement type for computing the volume' )


@twoArgFunctionEvaluator( )
@argValidator( [ MeasurementValidator( 0 ), IntValidator( 3 ) ] )
def getKSphereVolumeOperator( n, k ):
    return getKSphereVolume( n, k )


@oneArgFunctionEvaluator( )
@argValidator( [ MeasurementValidator( 0 ) ] )
def getSphereVolumeOperator( n ):
    return getKSphereVolume( n, 3 )


#******************************************************************************
#
#  getTriangleAreaOperator
#
#******************************************************************************

@argValidator( [ LengthValidator( ), LengthValidator( ), LengthValidator( ) ] )
def getTriangleAreaOperator( a, b, c ):
    s = divide( getSum( [ a, b, c ] ), 2 )   # semi-perimeter
    return getRoot( getProduct( [ s, subtract( s, a ), subtract( s, b ), subtract( s, c ) ] ), 2 )


#******************************************************************************
#
#  getTorusSurfaceAreaperator
#
#  http://preccalc.sourceforge.net/geometry.shtml
#
#  major = major radius
#  minor = minor radius
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ LengthValidator( ), LengthValidator( ) ] )
def getTorusSurfaceAreaOperator( major, minor ):
    return getProduct( [ 4, power( pi, 2 ), major, minor ] )


#******************************************************************************
#
#  getTorusVolumeOperator
#
#  http://preccalc.sourceforge.net/geometry.shtml
#
#  major = major radius
#  minor = minor radius
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ LengthValidator( ), LengthValidator( ) ] )
def getTorusVolumeOperator( major, minor ):
    return getProduct( [ 2, power( pi, 2 ), major, getPower( minor, 2 ) ] )


#******************************************************************************
#
#  getConeSurfaceAreaOperator
#
#  http://preccalc.sourceforge.net/geometry.shtml
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ LengthValidator( ), LengthValidator( ) ] )
def getConeSurfaceAreaOperator( radius, height ):
    hypotenuse = getRoot( add( getPower( radius, 2 ), getPower( height, 2 ) ), 2 )
    return getProduct( [ pi, radius, add( radius, hypotenuse ) ] )


#******************************************************************************
#
#  getConeVolumeOperator
#
#  http://preccalc.sourceforge.net/geometry.shtml
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ LengthValidator( ), LengthValidator( ) ] )
def getConeVolumeOperator( radius, height ):
    return getProduct( [ pi, getPower( radius, 2 ), divide( height, 3 ) ] )


#******************************************************************************
#
#  getTetrahedronSurfaceAreaOperator
#
#  https://en.wikipedia.org/wiki/Tetrahedron
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ LengthValidator( ) ] )
def getTetrahedronSurfaceAreaOperator( n ):
    return multiply( sqrt( 3 ), getPower( n, 2 ) )


#******************************************************************************
#
#  getTetrahedronVolumeOperator
#
#  https://en.wikipedia.org/wiki/Tetrahedron
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ LengthValidator( ) ] )
def getTetrahedronVolumeOperator( n ):
    return divide( getPower( n, 3 ), fmul( 6, sqrt( 2 ) ) )


#******************************************************************************
#
#  getOctahedronSurfaceAreaOperator
#
#  https://en.wikipedia.org/wiki/Octahedron
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ LengthValidator( ) ] )
def getOctahedronSurfaceAreaOperator( n ):
    return getProduct( [ 2, sqrt( 3 ), getPower( n, 2 ) ] )


#******************************************************************************
#
#  getOctahedronVolumeOperator
#
#  https://en.wikipedia.org/wiki/Octahedron
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ LengthValidator( ) ] )
def getOctahedronVolumeOperator( n ):
    return divide( multiply( sqrt( 2 ), getPower( n, 3 ) ), 3 )


#******************************************************************************
#
#  getDodecahedronSurfaceAreaOperator
#
#  https://en.wikipedia.org/wiki/Dodecahedron
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ LengthValidator( ) ] )
def getDodecahedronSurfaceAreaOperator( n ):
    area = getProduct( [ 3, getRoot( add( 25, fmul( 10, sqrt( 5 ) ) ), 2 ), getPower( n, 2 ) ] )
    return area.convert( 'meter^2' )


#******************************************************************************
#
#  getDodecahedronVolumeOperator
#
#  https://en.wikipedia.org/wiki/Dodecahedron
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ LengthValidator( ) ] )
def getDodecahedronVolumeOperator( n ):
    return divide( multiply( fadd( 15, fmul( 7, sqrt( 5 ) ) ), getPower( n, 3 ) ), 4 ).convert( 'meter^3' )


#******************************************************************************
#
#  getIcosahedronSurfaceAreaOperator
#
#  https://en.wikipedia.org/wiki/Icosahedron
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ LengthValidator( ) ] )
def getIcosahedronSurfaceAreaOperator( n ):
    return getProduct( [ 5, sqrt( 3 ), getPower( n, 2 ) ] ).convert( 'meter^2' )


#******************************************************************************
#
#  getIcosahedronVolumeOperator
#
#  https://en.wikipedia.org/wiki/Icosahedron
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ LengthValidator( ) ] )
def getIcosahedronVolumeOperator( n ):
    return getProduct( [ fdiv( 5, 12 ), fadd( 3, sqrt( 5 ) ), getPower( n, 3 ) ] ).convert( 'meter^3' )


#******************************************************************************
#
#  getAntiprismSurfaceAreaOperator
#
#  https://en.wikipedia.org/wiki/Antiprism
#
#  n = sides
#  k = edge length
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 3 ), LengthValidator( ) ] )
def getAntiprismSurfaceAreaOperator( n, k ):
    result = getProduct( [ fdiv( n, 2 ), fadd( cot( fdiv( pi, n ) ), sqrt( 3 ) ), getPower( k, 2 ) ] )
    return result.convert( 'meter^2' )


#******************************************************************************
#
#  getAntiprismVolumeOperator
#
#  http://www.fxsolver.com/browse/formulas/Antiprism+uniform+%28Volume%29
#
#  n = sides
#  k = edge length
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 3 ), LengthValidator( ) ] )
def getAntiprismVolumeOperator( n, k ):
    result = getProduct( [ fdiv( fprod( [ n, sqrt( fsub( fmul( 4, cos( cos( fdiv( pi, fmul( n, 2 ) ) ) ) ), 1 ) ),
                                          sin( fdiv( fmul( 3, pi ), fmul( 2, n ) ) ) ] ),
                                 fmul( 12, sin( sin( fdiv( pi, n ) ) ) ) ),
                           sin( fdiv( fmul( 3, pi ), fmul( 2, n ) ) ),
                           getPower( k, 3 ) ] )

    return result.convert( 'meter^3' )


#******************************************************************************
#
#  getPrismSurfaceAreaOperator
#
#  https://en.wikipedia.org/wiki/Prism
#
#  n = sides
#  k = edge length
#  h = height
#
#******************************************************************************

@argValidator( [ IntValidator( 3 ), LengthValidator( ), LengthValidator( ) ] )
def getPrismSurfaceAreaOperator( n, k, h ):
    result = add( getProduct( [ fdiv( n, 2 ), getPower( k, 2 ), cot( fdiv( pi, n ) ) ] ),
                  getProduct( [ n, k, h ] ) )
    return result.convert( 'meter^2' )


#******************************************************************************
#
#  getPrismVolumeOperator
#
#  https://en.wikipedia.org/wiki/Prism
#
#  n = sides
#  k = edge length
#  h = height
#
#******************************************************************************

@argValidator( [ IntValidator( 3 ), LengthValidator( ), LengthValidator( ) ] )
def getPrismVolumeOperator( n, k, h ):
    return getProduct( [ fdiv( n, 4 ), h, getPower( k, 2 ), cot( fdiv( pi, n ) ) ] ).convert( 'meter^3' )
